/* /////////////////////////////////////////////////////////////////////////////
// graphplan.c  -  15/05/2002 by aldeba
//
// Funcoes para a geracao do grafo de planejamento
///////////////////////////////////////////////////////////////////////////// */

/*	includes do fonte */

#include	<assert.h>
#include	<stdlib.h>
#include	<string.h>

#include	"graphplan.h"
#include	"memoria.h"

/* /////////////////////////////////////////////////////////////////////////////
// vertice do grafo de planejamento
///////////////////////////////////////////////////////////////////////////// */

/*
//	funcao para destruir um vetor de vertices de um grafo
*/

void destruirVertices( PVerticeGrafo _vetor )
	{
	assert( NULL != _vetor );

	SM_FREE( _vetor );
	}

/*
//	funcao para escrever um vetor de vertices de proposicoes de um grafo
//		em um stream
*/

void escreverVertices( PProblemaStrips _problema, long int _numVertices
		, PVerticeGrafo _vetor, char * _terminador, FILE * _stream )
	{
	long int	i = 0L;

	assert( NULL != _problema );
	assert( 0 < _numVertices );
	assert( NULL != _vetor );
	assert( NULL != _stream );

	for( ; _numVertices > i; i++ )
		{
		if( ACAO != _vetor[ i ].valorVerdade )
			{
/*	o vertice representa uma proposicao */

			assert( _problema->numProposicoes > _vetor[ i ].numProposicao );

			if( VERDADEIRO == _vetor[ i ].valorVerdade )
				fprintf( _stream, "%s"
						, _problema->proposicoes[ _vetor[ i ].numProposicao ].nome );
			else
				fprintf( _stream, "~%s"
						, _problema->proposicoes[ _vetor[ i ].numProposicao ].nome );
			}
		else
			{
/*	o vertice representa uma acao */

			assert( _problema->numOperadores > _vetor[ i ].numProposicao );

			if( 0 <= _vetor[ i ].numProposicao )
				fprintf( _stream, "%s"
						, _problema->dominio[ _vetor[ i ].numProposicao ].nome );
			else
				fprintf( _stream, "NOOP( %ld )", -1 * _vetor[ i ].numProposicao );
			}

		if( NULL != _terminador )
			fprintf( _stream, "%s", _terminador );
		}
	}

/*
//	funcao que retorna 1 se todas as proposicoes de uma vetor de estado pertencem
//		a uma vetor de vertices
*/

int estaContidoVertices( long int _numProposicoes, PEstado _vetorEstado
		, long int _numVertices, PVerticeGrafo _vetor )
	{
	long int	i = 0L;
	long int	numObjetivos = 0L;
	long int	numObjetivosSatisfeitos = 0L;

	assert( 0 < _numProposicoes );
	assert( NULL != _vetorEstado );
	assert( 0 < _numVertices );
	assert( NULL != _vetor );

//	conta o numero de proposicoes do estado objetivo

	for( i = 0L; _numProposicoes > i; i++ )
		{
		if( INDEFINIDO == _vetorEstado[ i ].valorVerdade )
			continue;

		numObjetivos++;
		}

//	conta o numero de proposicoes do estado objetivo presentes nos vertices

	for( i = 0L; _numVertices > i; i++ )
		{
		if( _vetor[ i ].valorVerdade
				== _vetorEstado[ _vetor[ i ].numProposicao ].valorVerdade )
			numObjetivosSatisfeitos++;
		}

	if( numObjetivos != numObjetivosSatisfeitos )
		return( 0 );

	return( 1 );
	}

/* /////////////////////////////////////////////////////////////////////////////
// aresta do grafo de planejamento
///////////////////////////////////////////////////////////////////////////// */

/*
//	funcao para destruir uma lista ligada de arestas de um grafo
*/

void destruirListaArestas( PArestaGrafo _lista )
	{
	PArestaGrafo	aresta;
	PArestaGrafo	auxiliar;

	assert( NULL != _lista );

	aresta = _lista;
	while( NULL != aresta )
		{
		auxiliar = aresta->proxima;

		SM_FREE( aresta );

		aresta = auxiliar;
		}
	}

/*
//	funcao para escrever uma lista ligada de arestas de um grafo em um stream
*/

void escreverListaArestas( PProblemaStrips _problema, PArestaGrafo _lista
		, char * _separador, FILE * _stream )
	{
	PArestaGrafo	aresta;

	assert( NULL != _problema );
	assert( NULL != _lista );
	assert( NULL != _stream );

	aresta = _lista;
	while( NULL != aresta )
		{
		fprintf( _stream, "\t" );
		escreverVertices( _problema, 1, aresta->vertice1, NULL, _stream );

		if( NULL == _separador )
			fprintf( _stream, " - " );
		else
			fprintf( _stream, "%s", _separador );

		escreverVertices( _problema, 1, aresta->vertice2, NULL, _stream );
		fprintf( _stream, "\n" );

		aresta = aresta->proxima;
		}
	}

/*
//	funcao para adicionar uma aresta a uma lista ligada de arestas
*/

PArestaGrafo adicionaAresta( PArestaGrafo _lista, PVerticeGrafo _vertice1
		, PVerticeGrafo _vertice2 )
	{
	PArestaGrafo	novaAresta;
	PArestaGrafo	ultimaAresta;

	assert( NULL != _vertice1 );
	assert( NULL != _vertice2 );

	ultimaAresta = _lista;
	while( NULL != ultimaAresta
			&& NULL != ultimaAresta->proxima )
		{
		ultimaAresta = ultimaAresta->proxima;
		}

	novaAresta = ( PArestaGrafo )SM_MALLOC( "TArestaGrafo"
			, sizeof( TArestaGrafo ) );
	if( NULL == novaAresta )
		{
		fprintf( stderr, "Erro na alocacao de memoria\n" );
		return( NULL );
		}

	novaAresta->vertice1 = _vertice1;
	novaAresta->vertice2 = _vertice2;
	novaAresta->proxima = NULL;

	if( NULL == ultimaAresta )
		return( novaAresta );

	ultimaAresta->proxima = novaAresta;

	return( _lista );
	}

/* /////////////////////////////////////////////////////////////////////////////
// nivel do grafo de planejamento
///////////////////////////////////////////////////////////////////////////// */

/*
//	funcao para contruir o nivel zero do grafo de planejamento
*/

PNivelGrafo construirNivelZero( PProblemaStrips _problema )
	{
	PNivelGrafo	nivelZero = NULL;
	long int	i = 0L;

	assert( NULL != _problema );

	nivelZero = ( PNivelGrafo )SM_MALLOC( "TNivelGrafo", sizeof( TNivelGrafo ) );
	if( NULL == nivelZero )
		{
		fprintf( stderr, "Erro na alocacao de memoria\n" );
		return( NULL );
		}

	nivelZero->numeroNivel = 0;
	nivelZero->arestaPreCondicao = NULL;
	nivelZero->numAcoes = 0L;
	nivelZero->verticeAcao = NULL;
	nivelZero->acoesExclusivas = NULL;
	nivelZero->arestaEfeito = NULL;
	nivelZero->numProposicoes = 0L;
	nivelZero->verticeProposicao = NULL;
	nivelZero->proposicoesExclusivas = NULL;
	nivelZero->proximo = NULL;

/*	aloca memoria para os vertices de proposicoes */

	nivelZero->verticeProposicao = ( PVerticeGrafo )SM_MALLOC( "TVerticeGrafo"
			, 2 * _problema->numProposicoes * sizeof( TVerticeGrafo ) );
	if( NULL == nivelZero->verticeProposicao )
		{
		fprintf( stderr, "Erro na alocacao de memoria\n" );
		return( NULL );
		}

/*	insere um vertice de proposicao para cada proposicao do estado inicial */

	for( ; _problema->numProposicoes > i; i++ )
		{
		if( INDEFINIDO == _problema->estadoInicial[ i ].valorVerdade )
			continue;

		nivelZero->verticeProposicao[ nivelZero->numProposicoes ].valorVerdade
				= _problema->estadoInicial[ i ].valorVerdade;
		nivelZero->verticeProposicao[ nivelZero->numProposicoes ].numProposicao
				= i;

		nivelZero->numProposicoes++;
		}

	return( nivelZero );
	}

/*
//	funcao para destruir um grafo de planejamento
*/

void destruirGrafoPlanejamento( PNivelGrafo _grafoPlanejamento )
	{
	PNivelGrafo	nivelGrafo;
	PNivelGrafo	auxiliar;

	assert( NULL != _grafoPlanejamento );

	nivelGrafo = _grafoPlanejamento;
	while( NULL != nivelGrafo )
		{
		auxiliar = nivelGrafo->proximo;

		if( NULL != nivelGrafo->arestaPreCondicao )
			destruirListaArestas( nivelGrafo->arestaPreCondicao );
		if( NULL != nivelGrafo->verticeAcao )
			destruirVertices( nivelGrafo->verticeAcao );
		if( NULL != nivelGrafo->acoesExclusivas )
			destruirListaArestas( nivelGrafo->acoesExclusivas );
		if( NULL != nivelGrafo->arestaEfeito )
			destruirListaArestas( nivelGrafo->arestaEfeito );
		if( NULL != nivelGrafo->verticeProposicao )
			destruirVertices( nivelGrafo->verticeProposicao );
		if( NULL != nivelGrafo->proposicoesExclusivas )
			destruirListaArestas( nivelGrafo->proposicoesExclusivas );

		SM_FREE( nivelGrafo );

		nivelGrafo = auxiliar;
		}
	}

/*
//	funcao para escrever um grafo de planejamento em um stream
*/

void escreverGrafoPlanejamento( PProblemaStrips _problema
		, PNivelGrafo _grafoPlanejamento, FILE * _stream )
	{
	PNivelGrafo	nivelGrafo;

	assert( NULL != _problema );
	assert( NULL != _grafoPlanejamento );
	assert( NULL != _stream );

	nivelGrafo = _grafoPlanejamento;
	while( NULL != nivelGrafo )
		{
		if( 0 == nivelGrafo->numeroNivel )
			{
			fprintf( _stream, ">>> vertices do nivel %d - "
					, nivelGrafo->numeroNivel );
			fprintf( _stream, "proposicoes do estado inicial\n" );

			escreverVertices( _problema, nivelGrafo->numProposicoes
					, nivelGrafo->verticeProposicao, "\n", _stream );
			}
		else
			{
			fprintf( _stream, ">>> vertices do nivel %d - "
					, nivelGrafo->numeroNivel - 1 );
			fprintf( _stream, "acoes aplicaveis no nivel %d\n"
					, nivelGrafo->numeroNivel - 2 );
			escreverVertices( _problema, nivelGrafo->numAcoes
					, nivelGrafo->verticeAcao, "\n", _stream );

			fprintf( _stream, ">>> arestas entre os niveis %d e %d"
					, nivelGrafo->numeroNivel - 2, nivelGrafo->numeroNivel - 1 );
			fprintf( _stream, " - arestas de pre-condicao\n" );
			escreverListaArestas( _problema, nivelGrafo->arestaPreCondicao, "->"
					, _stream );

			if( NULL != nivelGrafo->acoesExclusivas )
				{
				fprintf( _stream, ">>> mutex do nivel %d"
						, nivelGrafo->numeroNivel );
				fprintf( _stream, " - exclusao mutua entre acoes\n" );
				escreverListaArestas( _problema, nivelGrafo->acoesExclusivas
						, "||", _stream );
				}

			fprintf( _stream, ">>> vertices do nivel %d"
					, nivelGrafo->numeroNivel );
			fprintf( _stream, " - efeitos das acoes do nivel %d\n"
					, nivelGrafo->numeroNivel - 1 );

			escreverVertices( _problema, nivelGrafo->numProposicoes
					, nivelGrafo->verticeProposicao, "\n", _stream );
			}

		if( 1 < nivelGrafo->numeroNivel )
			{
			fprintf( _stream, ">>> arestas entre os niveis %d e %d"
					, nivelGrafo->numeroNivel - 1, nivelGrafo->numeroNivel );
			fprintf( _stream, " - arestas de efeito\n" );
			escreverListaArestas( _problema, nivelGrafo->arestaEfeito, "->"
					, _stream );

			if( NULL != nivelGrafo->proposicoesExclusivas )
				{
				fprintf( _stream, ">>> mutex do nivel %d"
						, nivelGrafo->numeroNivel );
				fprintf( _stream, " - exclusao mutua entre proposicoes\n" );
				escreverListaArestas( _problema
						, nivelGrafo->proposicoesExclusivas, "||", _stream );
				}
			}

		nivelGrafo = nivelGrafo->proximo;
		}
	}

/*
//	funcao para adicionar um nivel a um grafo de planejamento
*/

PNivelGrafo adicionaNivelGrafo( PProblemaStrips _problema
		, PNivelGrafo _grafoPlanejamento )
	{
	PNivelGrafo		ultimoNivel = NULL;
	PNivelGrafo		novoNivel = NULL;
	PEstado			efeito = NULL;
	PArestaGrafo	novaArestaPrecondicao;
	PArestaGrafo	novaArestaEfeito;
	long int		numProposicao;
	long int		i = 0;
	long int		j = 0;

	assert( NULL != _problema );
	assert( NULL != _grafoPlanejamento );

	ultimoNivel = _grafoPlanejamento;
	while( NULL != ultimoNivel->proximo )
		ultimoNivel = ultimoNivel->proximo;

/*	cria um estado auxiliar "vazio" */

	efeito = construirEstado( _problema->numProposicoes, _problema->proposicoes
			, "" );
	if( NULL == efeito )
		return NULL;

/*	aloca memoria para o novo nivel do grafo */

	novoNivel = ( PNivelGrafo )SM_MALLOC( "TNivelGrafo", sizeof( TNivelGrafo ) );
	if( NULL == novoNivel )
		{
		fprintf( stderr, "Erro na alocacao de memoria\n" );
		return( NULL );
		}

	novoNivel->numeroNivel = ultimoNivel->numeroNivel + 2;
	novoNivel->arestaPreCondicao = NULL;
	novoNivel->numAcoes = 0L;
	novoNivel->verticeAcao = NULL;
	novoNivel->acoesExclusivas = NULL;
	novoNivel->arestaEfeito = NULL;
	novoNivel->numProposicoes = 0L;
	novoNivel->verticeProposicao = NULL;
	novoNivel->proposicoesExclusivas = NULL;
	novoNivel->proximo = NULL;

	ultimoNivel->proximo = novoNivel;

/*	aloca memoria para os vertices de acoes e proposicoes */

	novoNivel->verticeAcao = ( PVerticeGrafo )SM_MALLOC( "TVerticeGrafo"
			, ( ultimoNivel->numProposicoes + _problema->numOperadores )
			* sizeof( TVerticeGrafo ) );
	novoNivel->verticeProposicao = ( PVerticeGrafo )SM_MALLOC( "TVerticeGrafo"
			, 2 * _problema->numProposicoes * sizeof( TVerticeGrafo ) );
	if( NULL == novoNivel->verticeAcao
			|| NULL == novoNivel->verticeProposicao )
		{
		fprintf( stderr, "Erro na alocacao de memoria\n" );
		return( NULL );
		}

/*	persiste com acoes NOP as proposicoes do ultimo nivel do grafo */

	for( i = 0; ultimoNivel->numProposicoes > i; i++ )
		{
		novaArestaPrecondicao = ( PArestaGrafo )SM_MALLOC( "TArestaGrafo"
				, sizeof( TArestaGrafo ) );
		novaArestaEfeito = ( PArestaGrafo )SM_MALLOC( "TArestaGrafo"
				, sizeof( TArestaGrafo ) );
		if( NULL == novaArestaPrecondicao
				|| NULL == novaArestaEfeito )
			{
			fprintf( stderr, "Erro na alocacao de memoria\n" );
			return( NULL );
			}

/*	insere um vertice com a acao NOP */

		novoNivel->verticeAcao[ novoNivel->numAcoes ].valorVerdade = ACAO;
		novoNivel->verticeAcao[ novoNivel->numAcoes ].numProposicao = -1 * ( i + 1 );
		novoNivel->verticeAcao[ novoNivel->numAcoes ].numeroPrecondicoes = 1;
		novoNivel->numAcoes++;

/*	insere um vertice com a proposicao persistida */

		novoNivel->verticeProposicao[ novoNivel->numProposicoes ].valorVerdade
				= ultimoNivel->verticeProposicao[ i ].valorVerdade;
		novoNivel->verticeProposicao[ novoNivel->numProposicoes ].numProposicao
				= ultimoNivel->verticeProposicao[ i ].numProposicao;
		novoNivel->numProposicoes++;

/*	insere uma aresta de precodicao */

		novaArestaPrecondicao->vertice1 = ultimoNivel->verticeProposicao + i;
		novaArestaPrecondicao->vertice2 = novoNivel->verticeAcao
				+ novoNivel->numAcoes - 1;
		if( NULL == novoNivel->arestaPreCondicao )
			novaArestaPrecondicao->proxima = NULL;
		else
			novaArestaPrecondicao->proxima = novoNivel->arestaPreCondicao;
		novoNivel->arestaPreCondicao = novaArestaPrecondicao;

/*	insere uma aresta de efeito */

		novaArestaEfeito->vertice1 = novoNivel->verticeAcao
				+ novoNivel->numAcoes - 1;
		novaArestaEfeito->vertice2 = novoNivel->verticeProposicao
				+ novoNivel->numProposicoes - 1;
		if( NULL == novoNivel->arestaEfeito )
			novaArestaEfeito->proxima = NULL;
		else
			novaArestaEfeito->proxima = novoNivel->arestaEfeito;
		novoNivel->arestaEfeito = novaArestaEfeito;
		}

/*	adiciona as acoes aplicaveis no ultimo nivel do grafo */

	for( i = 0; _problema->numOperadores > i; i++ )
		{
		if( 0 == estaContidoVertices( _problema->numProposicoes
				, _problema->dominio[ i ].precondicao
				, ultimoNivel->numProposicoes, ultimoNivel->verticeProposicao ) )
			continue;

/*	insere um vertice com a acao */

		novoNivel->verticeAcao[ novoNivel->numAcoes ].valorVerdade = ACAO;
		novoNivel->verticeAcao[ novoNivel->numAcoes ].numProposicao = i;
		novoNivel->verticeAcao[ novoNivel->numAcoes ].numeroPrecondicoes = 0;
		novoNivel->numAcoes++;

/*	insere as arestas de precondicao */

		for( j = 0; ultimoNivel->numProposicoes > j; j++ )
			{
			numProposicao = ultimoNivel->verticeProposicao[ j ].numProposicao;

			if( _problema->dominio[ i ].precondicao[ numProposicao ].valorVerdade
					!= ultimoNivel->verticeProposicao[ j ].valorVerdade )
				continue;

			novoNivel->verticeAcao[ novoNivel->numAcoes - 1 ].numeroPrecondicoes++;

			novaArestaPrecondicao = ( PArestaGrafo )SM_MALLOC( "TArestaGrafo"
					, sizeof( TArestaGrafo ) );
			if( NULL == novaArestaPrecondicao )
				{
				fprintf( stderr, "Erro na alocacao de memoria\n" );
				return( NULL );
				}

			novaArestaPrecondicao->vertice1 = ultimoNivel->verticeProposicao + j;
			novaArestaPrecondicao->vertice2 = novoNivel->verticeAcao
					+ novoNivel->numAcoes - 1;
			if( NULL == novoNivel->arestaPreCondicao )
				novaArestaPrecondicao->proxima = NULL;
			else
				novaArestaPrecondicao->proxima = novoNivel->arestaPreCondicao;
			novoNivel->arestaPreCondicao = novaArestaPrecondicao;
			}

/*	copia os efeitos da acao */

		for( j = 0; _problema->numProposicoes > j; j++ )
			efeito[ j ].valorVerdade = _problema->dominio[ i ].efeito[ j ].valorVerdade;

/*	insere as arestas de efeito para os vertices ja existentes no grafo */

		for( j = 0; novoNivel->numProposicoes > j; j++ )
			{
			numProposicao = novoNivel->verticeProposicao[ j ].numProposicao;

			if( efeito[ numProposicao ].valorVerdade
					!= novoNivel->verticeProposicao[ j ].valorVerdade )
				continue;

			novaArestaEfeito = ( PArestaGrafo )SM_MALLOC( "TArestaGrafo"
					, sizeof( TArestaGrafo ) );
			if( NULL == novaArestaEfeito )
				{
				fprintf( stderr, "Erro na alocacao de memoria\n" );
				return( NULL );
				}

			novaArestaEfeito->vertice1 = novoNivel->verticeAcao
					+ novoNivel->numAcoes - 1;
			novaArestaEfeito->vertice2 = novoNivel->verticeProposicao + j;
			if( NULL == novoNivel->arestaEfeito )
				novaArestaEfeito->proxima = NULL;
			else
				novaArestaEfeito->proxima = novoNivel->arestaEfeito;
			novoNivel->arestaEfeito = novaArestaEfeito;

/*	troca o valor verdade para marcar que esse efeito ja foi adicionado */

			efeito[ numProposicao ].valorVerdade = INDEFINIDO;
			}

/*	insere os novos vertices e arestas de efeito */

		for( j = 0; _problema->numProposicoes > j; j++ )
			{
			if( INDEFINIDO == efeito[ j ].valorVerdade )
				continue;

/*	adiciona um novo vertice de efeito */

			novoNivel->verticeProposicao[ novoNivel->numProposicoes ].valorVerdade
					= efeito[ j ].valorVerdade;
			novoNivel->verticeProposicao[ novoNivel->numProposicoes ].numProposicao = j;
			novoNivel->numProposicoes++;

/*	adiciona a aresta para o novo efeito */

			novaArestaEfeito = ( PArestaGrafo )SM_MALLOC( "TArestaGrafo"
					, sizeof( TArestaGrafo ) );
			if( NULL == novaArestaEfeito )
				{
				fprintf( stderr, "Erro na alocacao de memoria\n" );
				return( NULL );
				}

			novaArestaEfeito->vertice1 = novoNivel->verticeAcao
					+ novoNivel->numAcoes - 1;
			novaArestaEfeito->vertice2 = novoNivel->verticeProposicao
					+ novoNivel->numProposicoes - 1;
			if( NULL == novoNivel->arestaEfeito )
				novaArestaEfeito->proxima = NULL;
			else
				novaArestaEfeito->proxima = novoNivel->arestaEfeito;
			novoNivel->arestaEfeito = novaArestaEfeito;
			}
		}

	destruirEstado( efeito );

	return( _grafoPlanejamento );
	}

/*
//	funcao para adicionar as relacoes de exclusao mutua
//		entre as acoes do ultimo nivel do grafo de planejamento
*/

int adicionaArestasExclusaoMutuaAcoes( PNivelGrafo _grafoPlanejamento )
	{
	PNivelGrafo		penultimoNivel = NULL;
	PNivelGrafo		ultimoNivel = NULL;
	PArestaGrafo	aresta1;
	PArestaGrafo	aresta2;
	PArestaGrafo	relacaoExclusao;
	PArestaGrafo	novaArestaExclusao;
	int				exclusivas;
	long int		i = 0;
	long int		j = 0;

	assert( NULL != _grafoPlanejamento );

	ultimoNivel = _grafoPlanejamento;
	while( NULL != ultimoNivel->proximo )
		{
		penultimoNivel = ultimoNivel;
		ultimoNivel = ultimoNivel->proximo;
		}

/*	verifica todas as combinacoes de pares de acoes */

	for( i = 0; ultimoNivel->numAcoes > i; i++ )
		{
		for( j = 0; ultimoNivel->numAcoes > j; j++ )
			{
			if( i == j )
				continue;

			exclusivas = 0;

/*	verifica se ha efeitos conflitantes */

			aresta1 = ultimoNivel->arestaEfeito;
			while( NULL != aresta1 )
				{
				if( aresta1->vertice1 != ultimoNivel->verticeAcao + i )
					{
					aresta1 = aresta1->proxima;
					continue;
					}

				aresta2 = ultimoNivel->arestaEfeito;
				while( NULL != aresta2 )
					{
					if( aresta2->vertice1 != ultimoNivel->verticeAcao + j )
						{
						aresta2 = aresta2->proxima;
						continue;
						}

					if( aresta1->vertice2->numProposicao
							== aresta2->vertice2->numProposicao
							&& aresta1->vertice2->valorVerdade
							!= aresta2->vertice2->valorVerdade )
						{
						exclusivas = 1;
						break;
						}

					aresta2 = aresta2->proxima;
					}

				if( 1 == exclusivas )
					break;

				aresta1 = aresta1->proxima;
				}

/*	verifica se ha interferencia */

			if( 0 == exclusivas )
				{
				aresta1 = ultimoNivel->arestaPreCondicao;
				while( NULL != aresta1 )
					{
					if( aresta1->vertice2 != ultimoNivel->verticeAcao + i )
						{
						aresta1 = aresta1->proxima;
						continue;
						}

					aresta2 = ultimoNivel->arestaEfeito;
					while( NULL != aresta2 )
						{
						if( aresta2->vertice1 != ultimoNivel->verticeAcao + j )
							{
							aresta2 = aresta2->proxima;
							continue;
							}

						if( aresta1->vertice1->numProposicao
								== aresta2->vertice2->numProposicao
								&& aresta1->vertice1->valorVerdade
								!= aresta2->vertice2->valorVerdade )
							{
							exclusivas = 1;
							break;
							}

						aresta2 = aresta2->proxima;
						}

					if( 1 == exclusivas )
						break;

					aresta1 = aresta1->proxima;
					}
				}

/*	verifica se ha necessidades conflitantes */

			if( 0 == exclusivas
					&& NULL != penultimoNivel->proposicoesExclusivas )
				{
				aresta1 = ultimoNivel->arestaPreCondicao;
				while( NULL != aresta1 )
					{
					if( aresta1->vertice2 != ultimoNivel->verticeAcao + i )
						{
						aresta1 = aresta1->proxima;
						continue;
						}

					aresta2 = ultimoNivel->arestaPreCondicao;
					while( NULL != aresta2 )
						{
						if( aresta2->vertice2 != ultimoNivel->verticeAcao + j )
							{
							aresta2 = aresta2->proxima;
							continue;
							}

						if( aresta1->vertice1->numProposicao
								== aresta2->vertice1->numProposicao
								&& aresta1->vertice1->valorVerdade
								== aresta2->vertice1->valorVerdade )
							{
							aresta2 = aresta2->proxima;
							continue;
							}

						relacaoExclusao = penultimoNivel->proposicoesExclusivas;
						while( NULL != relacaoExclusao )
							{
							if( ( relacaoExclusao->vertice1 == aresta1->vertice1
									&& relacaoExclusao->vertice2 == aresta2->vertice1 )
									|| ( relacaoExclusao->vertice1 == aresta2->vertice1
									&& relacaoExclusao->vertice2 == aresta1->vertice1 ) )
								{
								exclusivas = 1;
								break;
								}

							relacaoExclusao = relacaoExclusao->proxima;
							}

						if( 1 == exclusivas )
							break;

						aresta2 = aresta2->proxima;
						}

					if( 1 == exclusivas )
						break;

					aresta1 = aresta1->proxima;
					}
				}

/*	adiciona uma aresta de exclusao mutua se necessario */

			if( 0 != exclusivas )
				{
				novaArestaExclusao = ( PArestaGrafo )SM_MALLOC( "TArestaGrafo"
						, sizeof( TArestaGrafo ) );
				if( NULL == novaArestaExclusao )
					{
					fprintf( stderr, "Erro na alocacao de memoria\n" );
					return( -1 );
					}

				novaArestaExclusao->vertice1 = ultimoNivel->verticeAcao + i;
				novaArestaExclusao->vertice2 = ultimoNivel->verticeAcao + j;
				if( NULL == ultimoNivel->acoesExclusivas )
					novaArestaExclusao->proxima = NULL;
				else
					novaArestaExclusao->proxima = ultimoNivel->acoesExclusivas;
				ultimoNivel->acoesExclusivas = novaArestaExclusao;
				}
			}
		}

	return( 0 );
	}

/*
//	funcao para adicionar as relacoes de exclusao mutua
//		entre as proposicoes do ultimo nivel do grafo de planejamento
*/

int adicionaArestasExclusaoMutuaProposicoes( PNivelGrafo _grafoPlanejamento )
	{
	PNivelGrafo		ultimoNivel = NULL;
	PArestaGrafo	aresta1;
	PArestaGrafo	aresta2;
	PArestaGrafo	relacaoExclusao;
	PArestaGrafo	novaArestaExclusao;
	int				exclusivas;
	long int		i = 0;
	long int		j = 0;

	assert( NULL != _grafoPlanejamento );

	ultimoNivel = _grafoPlanejamento;
	while( NULL != ultimoNivel->proximo )
		ultimoNivel = ultimoNivel->proximo;

/*	verifica todas as combinacoes de pares de proposicoes */

	for( i = 0; ultimoNivel->numProposicoes > i; i++ )
		{
		for( j = 0; ultimoNivel->numProposicoes > j; j++ )
			{
			if( i == j )
				continue;

			exclusivas = 1;

/*	verifica se todas as acoes que adicionam p1 */
/*		sao exclusivas a todas as acoes que adicionam p2 */

			aresta1 = ultimoNivel->arestaEfeito;
			while( NULL != aresta1 )
				{
				if( aresta1->vertice2 != ultimoNivel->verticeProposicao + i )
					{
					aresta1 = aresta1->proxima;
					continue;
					}

				aresta2 = ultimoNivel->arestaEfeito;
				while( NULL != aresta2 )
					{
					if( aresta2->vertice2 != ultimoNivel->verticeProposicao + j )
						{
						aresta2 = aresta2->proxima;
						continue;
						}

					if( aresta1->vertice1 == aresta2->vertice1 )
						{
						exclusivas = 0;
						break;
						}

					relacaoExclusao = ultimoNivel->acoesExclusivas;
					for( ;; )
						{
						if( NULL == relacaoExclusao )
							{
							exclusivas = 0;
							break;
							}

						if( ( relacaoExclusao->vertice1 == aresta1->vertice1
								&& relacaoExclusao->vertice2 == aresta2->vertice1 )
								|| ( relacaoExclusao->vertice1 == aresta2->vertice1
								&& relacaoExclusao->vertice2 == aresta1->vertice1 ) )
							break;

						relacaoExclusao = relacaoExclusao->proxima;
						}

					if( 0 == exclusivas )
						break;

					aresta2 = aresta2->proxima;
					}

				if( 0 == exclusivas )
					break;

				aresta1 = aresta1->proxima;
				}

/*	adiciona uma aresta de exclusao mutua se necessario */

			if( 0 != exclusivas )
				{
				novaArestaExclusao = ( PArestaGrafo )SM_MALLOC( "TArestaGrafo"
						, sizeof( TArestaGrafo ) );
				if( NULL == novaArestaExclusao )
					{
					fprintf( stderr, "Erro na alocacao de memoria\n" );
					return( -1 );
					}

				novaArestaExclusao->vertice1 = ultimoNivel->verticeProposicao + i;
				novaArestaExclusao->vertice2 = ultimoNivel->verticeProposicao + j;
				if( NULL == ultimoNivel->proposicoesExclusivas )
					novaArestaExclusao->proxima = NULL;
				else
					novaArestaExclusao->proxima = ultimoNivel->proposicoesExclusivas;
				ultimoNivel->proposicoesExclusivas = novaArestaExclusao;
				}
			}
		}

	return( 0 );
	}

/*
//	funcao que retorna 1 se todas as proposicao do estado objetivo do problema
//		esta presenta na lista de vertices de proposicao do ultimo nivel do grafo
*/

int objetivoSatisfeito( PProblemaStrips _problema
		, PNivelGrafo _grafoPlanejamento )
	{
	PNivelGrafo		ultimoNivel = NULL;
	PArestaGrafo	relacaoExclusao;
	long int		i = 0;
	long int		j = 0;

	assert( NULL != _problema );
	assert( NULL != _grafoPlanejamento );

	ultimoNivel = _grafoPlanejamento;
	while( NULL != ultimoNivel->proximo )
		ultimoNivel = ultimoNivel->proximo;

	if( 0 == estaContidoVertices( _problema->numProposicoes, _problema->objetivo
			, ultimoNivel->numProposicoes, ultimoNivel->verticeProposicao ) )
		return( 0 );

/*	verifica todas as combinacoes de pares de proposicoes */

	if( NULL == ultimoNivel->proposicoesExclusivas )
		return( 1 );

	for( i = 0; _problema->numProposicoes > i; i++ )
		{
		for( j = 0; _problema->numProposicoes > j; j++ )
			{
			if( i == j )
				continue;

			if( INDEFINIDO == _problema->objetivo[ i ].valorVerdade
					|| INDEFINIDO == _problema->objetivo[ j ].valorVerdade )
				continue;

			relacaoExclusao = ultimoNivel->proposicoesExclusivas;
			while( NULL != relacaoExclusao )
				{
				if( relacaoExclusao->vertice1->numProposicao == i
						&& relacaoExclusao->vertice1->valorVerdade
						== _problema->objetivo[ i ].valorVerdade

						&& relacaoExclusao->vertice2->numProposicao == j
						&& relacaoExclusao->vertice2->valorVerdade
						== _problema->objetivo[ j ].valorVerdade )
					return( 0 );

				if( relacaoExclusao->vertice1->numProposicao == j
						&& relacaoExclusao->vertice1->valorVerdade
						== _problema->objetivo[ j ].valorVerdade

						&& relacaoExclusao->vertice2->numProposicao == i
						&& relacaoExclusao->vertice2->valorVerdade
						== _problema->objetivo[ i ].valorVerdade )
					return( 0 );

				relacaoExclusao = relacaoExclusao->proxima;
				}
			}
		}

	return( 1 );
	}

/*
//	funcao para contruir o grafo de planejamento a partir de um problema STRIPS
*/

PNivelGrafo construirGrafoPlanejamento( PProblemaStrips _problema
		, int _condicoesParada )
	{
	PNivelGrafo	grafoPlanejamento = NULL;
	int			condicoesParadaObtidas = 0;

	assert( NULL != _problema );

	grafoPlanejamento = construirNivelZero( _problema );
	if( NULL == grafoPlanejamento )
		return( NULL );

	for( ;; )
		{
		if( 1 == objetivoSatisfeito( _problema, grafoPlanejamento ) )
			condicoesParadaObtidas++;

		if( condicoesParadaObtidas == _condicoesParada )
			break;

		grafoPlanejamento = adicionaNivelGrafo( _problema, grafoPlanejamento );
		if( NULL == grafoPlanejamento )
			return( NULL );

		if( adicionaArestasExclusaoMutuaAcoes( grafoPlanejamento ) != 0 )
			return( NULL );

		if( adicionaArestasExclusaoMutuaProposicoes( grafoPlanejamento ) != 0 )
			return( NULL );

#if		!defined( NDEBUG )

		escreverGrafoPlanejamento( _problema, grafoPlanejamento, stderr );
#endif	/* !defined( NDEBUG ) */
		}

	return( grafoPlanejamento );
	}

/* /////////////////////////////////////////////////////////////////////////////
// escolha de acao do grafo de planejamento
///////////////////////////////////////////////////////////////////////////// */

/*
//	funcao para destruir uma escolha de acoes
*/

void destruirEscolhaAcoes( PEscolhaAcao _planoSolucao )
	{
	PEscolhaAcao	escolhaAcao;
	PEscolhaAcao	auxiliar;

	assert( NULL != _planoSolucao );

	escolhaAcao = _planoSolucao;
	while( NULL != escolhaAcao )
		{
		auxiliar = escolhaAcao->proxima;

		SM_FREE( escolhaAcao );

		escolhaAcao = auxiliar;
		}
	}

/*
//	funcao para escrever uma escolha de acoes em um stream
*/

void escreverEscolhaAcoes( PProblemaStrips _problema, PEscolhaAcao _planoSolucao
		, FILE * _stream )
	{
	PEscolhaAcao	acaoPlano;

	assert( NULL != _problema );
	assert( NULL != _stream );

	acaoPlano = _planoSolucao;
	while( NULL != acaoPlano )
		{
		fprintf( _stream, "%s", _problema->dominio[ acaoPlano->acaoEscolhida ].nome );
		fprintf( _stream, "( %d )\n", acaoPlano->numeroNivel );

		acaoPlano = acaoPlano->proxima;
		}
	}

/*
//	funcao para extrair uma solucao do grafo de planejamento
*/

PEscolhaAcao extrairSolucao( PProblemaStrips _problema
		, PNivelGrafo _grafoPlanejamento )
	{
	PEscolhaAcao		planoSolucao = NULL;
/*
	PEscolhaAcao		novaEscolhaAcao;
	PNivelGrafo			ultimoNivel = NULL;
	PNivelGrafo			nivelAnterior;
	PListaProposicoes	objetivoNivel = NULL;
	PListaProposicoes	objetivoNivelAnterior = NULL;
	PListaProposicoes	proposicaoObjetivo;
	PListaProposicoes	novaProposicao;
	PVerticeGrafo		proposicao;
	PArestaGrafo		arestaEfeito;
	PArestaGrafo		arestaPreCondicao;

	assert( NULL != _grafoPlanejamento );
	assert( NULL != _objetivo );

	ultimoNivel = _grafoPlanejamento;
	while( NULL != ultimoNivel->proximo )
		ultimoNivel = ultimoNivel->proximo;
*/
/*	FALTA - transformar essa implementacao para grafos normais */

/*	copia no estado objetivo do nivel as proposicoes do estado objetivo */
/*
	proposicaoObjetivo = _objetivo;
	while( NULL != proposicaoObjetivo )
		{
		novaProposicao = ( PListaProposicoes )SM_MALLOC( "TListaProposicoes"
				, sizeof( TListaProposicoes ) );
		if( NULL == novaProposicao )
			{
			fprintf( stderr, "Erro na alocacao de memoria\n" );
			return( NULL );
			}

		strcpy( novaProposicao->proposicao, proposicaoObjetivo->proposicao );
		if( NULL == objetivoNivel )
			novaProposicao->proxima = NULL;
		else
			novaProposicao->proxima = objetivoNivel;
		objetivoNivel = novaProposicao;

		proposicaoObjetivo = proposicaoObjetivo->proxima;
		}
*/
/*	percorre o grafo de planejamento do ultimo nivel para o primeiro */
/*
	while( ultimoNivel != _grafoPlanejamento )
		{
*/
/*	percorre cada proposicao do objetivo do nivel */
/*
		proposicaoObjetivo = objetivoNivel;
		while( NULL != proposicaoObjetivo )
			{
			proposicao = ultimoNivel->verticeProposicao;
			while( NULL != proposicao )
				{
				if( 0 == strcmp( proposicao->proposicao
						, proposicaoObjetivo->proposicao ) )
					break;

				proposicao = proposicao->proximo;
				}

			assert( NULL != proposicao );
*/
/*	procura inicialmente uma acao NOOP */
/*
			arestaEfeito = ultimoNivel->arestaEfeito;
			while( NULL != arestaEfeito )
				{
				if( arestaEfeito->vertice2 == proposicao
						&& 0 == strncmp( arestaEfeito->vertice1->proposicao
						, "NOOP-", 5 ) )
					break;

				arestaEfeito = arestaEfeito->proxima;
				}
*/
/*	procura uma acao */
/*
			if( NULL == arestaEfeito )
				{
				arestaEfeito = ultimoNivel->arestaEfeito;
				while( NULL != arestaEfeito )
					{
					if( arestaEfeito->vertice2 == proposicao )
						break;

					arestaEfeito = arestaEfeito->proxima;
					}

				assert( NULL != arestaEfeito );
*/
/*	adiciona a acao escolhida na solucao */
/*
				novaEscolhaAcao = ( PEscolhaAcao )SM_MALLOC( "TEscolhaAcao"
						, sizeof( TEscolhaAcao ) );
				if( NULL == novaEscolhaAcao )
					{
					fprintf( stderr, "Erro na alocacao de memoria\n" );
					return( NULL );
					}

				novaEscolhaAcao->nivelAcao = ultimoNivel;
				novaEscolhaAcao->acaoEscolhida = arestaEfeito->vertice1;
				if( NULL == planoSolucao )
					novaEscolhaAcao->proxima = NULL;
				else
					novaEscolhaAcao->proxima = planoSolucao;
				planoSolucao = novaEscolhaAcao;
				}
*/
/*	adiciona pre-condicoes da acao escolhida aos objetivos do nivel anterior */
/*
			arestaPreCondicao = ultimoNivel->arestaPreCondicao;
			while( NULL != arestaPreCondicao )
				{
				if( arestaPreCondicao->vertice2 == arestaEfeito->vertice1 )
					{
					novaProposicao = ( PListaProposicoes )SM_MALLOC( "TListaProposicoes"
							, sizeof( TListaProposicoes ) );
					if( NULL == novaProposicao )
						{
						fprintf( stderr, "Erro na alocacao de memoria\n" );
						return( NULL );
						}

					strcpy( novaProposicao->proposicao
							, arestaPreCondicao->vertice1->proposicao );
					if( NULL == objetivoNivelAnterior )
						novaProposicao->proxima = NULL;
					else
						novaProposicao->proxima = objetivoNivelAnterior;
					objetivoNivelAnterior = novaProposicao;
					}

				arestaPreCondicao = arestaPreCondicao->proxima;
				}

			proposicaoObjetivo = proposicaoObjetivo->proxima;
			}
*/
/*	passa para o nivel anterior do grafo */
/*
		nivelAnterior = _grafoPlanejamento;
		while( ultimoNivel != nivelAnterior->proximo )
			nivelAnterior = nivelAnterior->proximo;
*/
/*	FALTA - aqui ha muita memoria sendo perdida (leaks) */
/*
		ultimoNivel = nivelAnterior;
		objetivoNivel = objetivoNivelAnterior;
		objetivoNivelAnterior = NULL;
		}
*/
	return( planoSolucao );
	}

