/* /////////////////////////////////////////////////////////////////////////////
// graphplan.h  -  18/10/2002 by aldeba
//
// Estruturas e prototipos do algoritmo para a geracao do grafo de planejamento
///////////////////////////////////////////////////////////////////////////// */

/*	protecao contra referencias circulares */

#if			!defined( __GRAPHPLAN_H )

#define		__GRAPHPLAN_H

/*	includes do fonte */

#include	"strips.h"

/* /////////////////////////////////////////////////////////////////////////////
// declaracoes de estruturas de dados
///////////////////////////////////////////////////////////////////////////// */

//	vertice do grafo de planejamento

#define	ACAO	0

struct verticeGrafo {
	short int				valorVerdade;
	long int				numProposicao;
	int						numeroPrecondicoes;
};

typedef struct verticeGrafo	TVerticeGrafo;
typedef TVerticeGrafo *		PVerticeGrafo;

//	aresta do grafo de planejamento

struct arestaGrafo {
	PVerticeGrafo			vertice1;
	PVerticeGrafo			vertice2;
	struct arestaGrafo *	proxima;
};

typedef struct arestaGrafo	TArestaGrafo;
typedef TArestaGrafo *		PArestaGrafo;

//	nivel do grafo de planejamento

struct nivelGrafo {
	int						numeroNivel;
	PArestaGrafo			arestaPreCondicao;
	long int				numAcoes;
	PVerticeGrafo			verticeAcao;
	PArestaGrafo			acoesExclusivas;
	PArestaGrafo			arestaEfeito;
	long int				numProposicoes;
	PVerticeGrafo			verticeProposicao;
	PArestaGrafo			proposicoesExclusivas;
	struct nivelGrafo *		proximo;
};

typedef struct nivelGrafo	TNivelGrafo;
typedef TNivelGrafo *		PNivelGrafo;

//	escolha de acao do grafo de planejamento

struct escolhaAcao {
	int						numeroNivel;
	long int				acaoEscolhida;
	struct escolhaAcao *	proxima;
};

typedef struct escolhaAcao	TEscolhaAcao;
typedef TEscolhaAcao *		PEscolhaAcao;

/* /////////////////////////////////////////////////////////////////////////////
// declaracoes de prototipos
///////////////////////////////////////////////////////////////////////////// */

void			destruirVertices						( PVerticeGrafo _vetor );
void			escreverVertices						( PProblemaStrips _problema, long int _numVertices, PVerticeGrafo _vetor, char * _terminador, FILE * _stream );
int				estaContidoVertices						( long int _numProposicoes, PEstado _vetorEstado, long int _numVertices, PVerticeGrafo _vetor );

void			destruirListaArestas					( PArestaGrafo _lista );
void			escreverListaArestas					( PProblemaStrips _problema, PArestaGrafo _lista, char * _separador, FILE * _stream );
PArestaGrafo	adicionaAresta							( PArestaGrafo _lista, PVerticeGrafo _vertice1, PVerticeGrafo _vertice2 );

PNivelGrafo		construirNivelZero						( PProblemaStrips _problema );
void			destruirGrafoPlanejamento				( PNivelGrafo _grafoPlanejamento );
void			escreverGrafoPlanejamento				( PProblemaStrips _problema, PNivelGrafo _grafoPlanejamento, FILE * _stream );
PNivelGrafo		adicionaNivelGrafo						( PProblemaStrips _problema, PNivelGrafo _grafoPlanejamento );
int				adicionaArestasExclusaoMutuaAcoes		( PNivelGrafo _grafoPlanejamento );
int				adicionaArestasExclusaoMutuaProposicoes	( PNivelGrafo _grafoPlanejamento );
int				objetivoSatisfeito						( PProblemaStrips _problema, PNivelGrafo _grafoPlanejamento );
PNivelGrafo		construirGrafoPlanejamento				( PProblemaStrips _problema, int _condicoesParada );

void			destruirEscolhaAcoes					( PEscolhaAcao _planoSolucao );
void			escreverEscolhaAcoes					( PProblemaStrips _problema, PEscolhaAcao _planoSolucao, FILE * _stream );
#endif		/* !defined( __GRAPHPLAN_H ) */

