/* /////////////////////////////////////////////////////////////////////////////
// graphplanRelaxado.c  -  18/10/2002 by aldeba
//
// Funcoes para a geracao do grafo de planejamento relaxado
///////////////////////////////////////////////////////////////////////////// */

/*	includes do fonte */

#include	<assert.h>
#include	<stdlib.h>
#include	<string.h>

#include	"graphplanRelaxado.h"
#include	"memoria.h"

/* /////////////////////////////////////////////////////////////////////////////
// nivel do grafo de planejamento
///////////////////////////////////////////////////////////////////////////// */

/*
//	funcao para adicionar um nivel a um grafo de planejamento relaxado
*/

PNivelGrafo adicionaNivelGrafoRelaxado( PProblemaStrips _problema
		, PNivelGrafo _grafoPlanejamento )
	{
	PNivelGrafo		ultimoNivel = NULL;
	PNivelGrafo		novoNivel = NULL;
	PEstado			efeito = NULL;
	PArestaGrafo	novaArestaPrecondicao;
	PArestaGrafo	novaArestaEfeito;
	long int		numProposicao;
	long int		i = 0;
	long int		j = 0;

	assert( NULL != _problema );
	assert( NULL != _grafoPlanejamento );

	ultimoNivel = _grafoPlanejamento;
	while( NULL != ultimoNivel->proximo )
		ultimoNivel = ultimoNivel->proximo;

/*	cria um estado auxiliar "vazio" */

	efeito = construirEstado( _problema->numProposicoes, _problema->proposicoes
			, "" );
	if( NULL == efeito )
		return NULL;

/*	aloca memoria para o novo nivel do grafo */

	novoNivel = ( PNivelGrafo )SM_MALLOC( "TNivelGrafo", sizeof( TNivelGrafo ) );
	if( NULL == novoNivel )
		{
		fprintf( stderr, "Erro na alocacao de memoria\n" );
		return( NULL );
		}

	novoNivel->numeroNivel = ultimoNivel->numeroNivel + 2;
	novoNivel->arestaPreCondicao = NULL;
	novoNivel->numAcoes = 0L;
	novoNivel->verticeAcao = NULL;
	novoNivel->acoesExclusivas = NULL;
	novoNivel->arestaEfeito = NULL;
	novoNivel->numProposicoes = 0L;
	novoNivel->verticeProposicao = NULL;
	novoNivel->proposicoesExclusivas = NULL;
	novoNivel->proximo = NULL;

	ultimoNivel->proximo = novoNivel;

/*	aloca memoria para os vertices de acoes e proposicoes */

	novoNivel->verticeAcao = ( PVerticeGrafo )SM_MALLOC( "TVerticeGrafo"
			, ( ultimoNivel->numProposicoes + _problema->numOperadores )
			* sizeof( TVerticeGrafo ) );
	novoNivel->verticeProposicao = ( PVerticeGrafo )SM_MALLOC( "TVerticeGrafo"
			, 2 * _problema->numProposicoes * sizeof( TVerticeGrafo ) );
	if( NULL == novoNivel->verticeAcao
			|| NULL == novoNivel->verticeProposicao )
		{
		fprintf( stderr, "Erro na alocacao de memoria\n" );
		return( NULL );
		}

/*	persiste com acoes NOP as proposicoes do ultimo nivel do grafo */

	for( i = 0; ultimoNivel->numProposicoes > i; i++ )
		{
		novaArestaPrecondicao = ( PArestaGrafo )SM_MALLOC( "TArestaGrafo"
				, sizeof( TArestaGrafo ) );
		novaArestaEfeito = ( PArestaGrafo )SM_MALLOC( "TArestaGrafo"
				, sizeof( TArestaGrafo ) );
		if( NULL == novaArestaPrecondicao
				|| NULL == novaArestaEfeito )
			{
			fprintf( stderr, "Erro na alocacao de memoria\n" );
			return( NULL );
			}

/*	insere um vertice com a acao NOP */

		novoNivel->verticeAcao[ novoNivel->numAcoes ].valorVerdade = ACAO;
		novoNivel->verticeAcao[ novoNivel->numAcoes ].numProposicao = -1 * ( i + 1 );
		novoNivel->verticeAcao[ novoNivel->numAcoes ].numeroPrecondicoes = 1;
		novoNivel->numAcoes++;

/*	insere um vertice com a proposicao persistida */

		novoNivel->verticeProposicao[ novoNivel->numProposicoes ].valorVerdade
				= ultimoNivel->verticeProposicao[ i ].valorVerdade;
		novoNivel->verticeProposicao[ novoNivel->numProposicoes ].numProposicao
				= ultimoNivel->verticeProposicao[ i ].numProposicao;
		novoNivel->numProposicoes++;

/*	insere uma aresta de precodicao */

		novaArestaPrecondicao->vertice1 = ultimoNivel->verticeProposicao + i;
		novaArestaPrecondicao->vertice2 = novoNivel->verticeAcao
				+ novoNivel->numAcoes - 1;
		if( NULL == novoNivel->arestaPreCondicao )
			novaArestaPrecondicao->proxima = NULL;
		else
			novaArestaPrecondicao->proxima = novoNivel->arestaPreCondicao;
		novoNivel->arestaPreCondicao = novaArestaPrecondicao;

/*	insere uma aresta de efeito */

		novaArestaEfeito->vertice1 = novoNivel->verticeAcao
				+ novoNivel->numAcoes - 1;
		novaArestaEfeito->vertice2 = novoNivel->verticeProposicao
				+ novoNivel->numProposicoes - 1;
		if( NULL == novoNivel->arestaEfeito )
			novaArestaEfeito->proxima = NULL;
		else
			novaArestaEfeito->proxima = novoNivel->arestaEfeito;
		novoNivel->arestaEfeito = novaArestaEfeito;
		}

/*	adiciona as acoes aplicaveis no ultimo nivel do grafo */

	for( i = 0; _problema->numOperadores > i; i++ )
		{
		if( 0 == estaContidoVertices( _problema->numProposicoes
				, _problema->dominio[ i ].precondicao
				, ultimoNivel->numProposicoes, ultimoNivel->verticeProposicao ) )
			continue;

/*	insere um vertice com a acao */

		novoNivel->verticeAcao[ novoNivel->numAcoes ].valorVerdade = ACAO;
		novoNivel->verticeAcao[ novoNivel->numAcoes ].numProposicao = i;
		novoNivel->verticeAcao[ novoNivel->numAcoes ].numeroPrecondicoes = 0;
		novoNivel->numAcoes++;

/*	insere as arestas de precondicao */

		for( j = 0; ultimoNivel->numProposicoes > j; j++ )
			{
			numProposicao = ultimoNivel->verticeProposicao[ j ].numProposicao;

			if( _problema->dominio[ i ].precondicao[ numProposicao ].valorVerdade
					!= ultimoNivel->verticeProposicao[ j ].valorVerdade )
				continue;

			novoNivel->verticeAcao[ novoNivel->numAcoes - 1 ].numeroPrecondicoes++;

			novaArestaPrecondicao = ( PArestaGrafo )SM_MALLOC( "TArestaGrafo"
					, sizeof( TArestaGrafo ) );
			if( NULL == novaArestaPrecondicao )
				{
				fprintf( stderr, "Erro na alocacao de memoria\n" );
				return( NULL );
				}

			novaArestaPrecondicao->vertice1 = ultimoNivel->verticeProposicao + j;
			novaArestaPrecondicao->vertice2 = novoNivel->verticeAcao
					+ novoNivel->numAcoes - 1;
			if( NULL == novoNivel->arestaPreCondicao )
				novaArestaPrecondicao->proxima = NULL;
			else
				novaArestaPrecondicao->proxima = novoNivel->arestaPreCondicao;
			novoNivel->arestaPreCondicao = novaArestaPrecondicao;
			}

/*	copia os efeitos da acao */

		copiarEstado( _problema->numProposicoes, efeito
				, _problema->dominio[ i ].efeito );

/*	insere as arestas de efeito para os vertices ja existentes no grafo */

		for( j = 0; novoNivel->numProposicoes > j; j++ )
			{
			numProposicao = novoNivel->verticeProposicao[ j ].numProposicao;

/*	para a tarefa relaxada, ignora os efeitos negativos */

			if( FALSO == efeito[ numProposicao ].valorVerdade )
				continue;

			if( efeito[ numProposicao ].valorVerdade
					!= novoNivel->verticeProposicao[ j ].valorVerdade )
				continue;

			novaArestaEfeito = ( PArestaGrafo )SM_MALLOC( "TArestaGrafo"
					, sizeof( TArestaGrafo ) );
			if( NULL == novaArestaEfeito )
				{
				fprintf( stderr, "Erro na alocacao de memoria\n" );
				return( NULL );
				}

			novaArestaEfeito->vertice1 = novoNivel->verticeAcao
					+ novoNivel->numAcoes - 1;
			novaArestaEfeito->vertice2 = novoNivel->verticeProposicao + j;
			if( NULL == novoNivel->arestaEfeito )
				novaArestaEfeito->proxima = NULL;
			else
				novaArestaEfeito->proxima = novoNivel->arestaEfeito;
			novoNivel->arestaEfeito = novaArestaEfeito;

/*	troca o valor verdade para marcar que esse efeito ja foi adicionado */

			efeito[ numProposicao ].valorVerdade = INDEFINIDO;
			}

/*	insere os novos vertices e arestas de efeito */

		for( j = 0; _problema->numProposicoes > j; j++ )
			{
			if( INDEFINIDO == efeito[ j ].valorVerdade )
				continue;

/*	para a tarefa relaxada, ignora os efeitos negativos */

			if( FALSO == efeito[ j ].valorVerdade )
				continue;

/*	adiciona um novo vertice de efeito */

			novoNivel->verticeProposicao[ novoNivel->numProposicoes ].valorVerdade
					= efeito[ j ].valorVerdade;
			novoNivel->verticeProposicao[ novoNivel->numProposicoes ].numProposicao = j;
			novoNivel->numProposicoes++;

/*	adiciona a aresta para o novo efeito */

			novaArestaEfeito = ( PArestaGrafo )SM_MALLOC( "TArestaGrafo"
					, sizeof( TArestaGrafo ) );
			if( NULL == novaArestaEfeito )
				{
				fprintf( stderr, "Erro na alocacao de memoria\n" );
				return( NULL );
				}

			novaArestaEfeito->vertice1 = novoNivel->verticeAcao
					+ novoNivel->numAcoes - 1;
			novaArestaEfeito->vertice2 = novoNivel->verticeProposicao
					+ novoNivel->numProposicoes - 1;
			if( NULL == novoNivel->arestaEfeito )
				novaArestaEfeito->proxima = NULL;
			else
				novaArestaEfeito->proxima = novoNivel->arestaEfeito;
			novoNivel->arestaEfeito = novaArestaEfeito;
			}
		}

	destruirEstado( efeito );

	return( _grafoPlanejamento );
	}

/*
//	funcao para construir um grafo de planejamento relaxado a partir de um
//		problema STRIPS
*/

PNivelGrafo construirGrafoRelaxado( PProblemaStrips _problema )
	{
	PNivelGrafo	grafoPlanejamento = NULL;
	PNivelGrafo	ultimoNivel = NULL;
	int			acoesNivelAnterior;
	int			acoesUltimoNivel;
	int			numeroNivel = 0;

	assert( NULL != _problema );

	grafoPlanejamento = construirNivelZero( _problema );
	if( NULL == grafoPlanejamento )
		return( NULL );

	for( ;; )
		{
		if( 1 == objetivoSatisfeito( _problema, grafoPlanejamento ) )
			{
#if		!defined( NDEBUG )

			fprintf( stderr, "Ultimo nivel do grafo: %d\n", numeroNivel );
#endif	/* !defined( NDEBUG ) */
			break;
			}

		grafoPlanejamento = adicionaNivelGrafoRelaxado( _problema
				, grafoPlanejamento );
		if( NULL == grafoPlanejamento )
			return( NULL );

/*	verifica se atingiu um ponto fixo */

		acoesNivelAnterior = acoesUltimoNivel = 0;

		ultimoNivel = grafoPlanejamento;
		while( NULL != ultimoNivel->proximo )
			{
			acoesNivelAnterior = ultimoNivel->numAcoes;

			ultimoNivel = ultimoNivel->proximo;
			acoesUltimoNivel = ultimoNivel->numAcoes;
			}

		if( 0 == acoesUltimoNivel - acoesNivelAnterior )
			{
#if		!defined( NDEBUG )

			fprintf( stderr, "Atingiu um ponto fixo\n" );
#endif	/* !defined( NDEBUG ) */

			destruirGrafoPlanejamento( grafoPlanejamento );
			return( NULL );
			}

		numeroNivel += 2;
		}

	return( grafoPlanejamento );
	}

/* /////////////////////////////////////////////////////////////////////////////
// escolha de acao do grafo de planejamento
///////////////////////////////////////////////////////////////////////////// */

/*
//	funcao para extrair uma solucao relaxada do grafo de planejamento
*/

long int extrairSolucaoRelaxada( PProblemaStrips _problema
		, PNivelGrafo _grafoPlanejamento, PEscolhaAcao _planoRelaxado )
	{
//	PEscolhaAcao	planoSolucao = NULL;
//	PEscolhaAcao	novaEscolhaAcao;
//	PEscolhaAcao	acaoPlano;
	PNivelGrafo		ultimoNivel = NULL;
	PNivelGrafo		nivelAnterior;
	PEstado			objetivoNivel = NULL;
	PEstado			objetivoNivelAnterior = NULL;
	PArestaGrafo	arestaEfeito;
	PArestaGrafo	arestaEscolhida;
	PArestaGrafo	arestaPreCondicao;
	long int		numeroAcoes = 0L;
	long int		i = 0L;
	long int		j = 0L;
	long int		k = 0L;

	assert( NULL != _problema );
	assert( NULL != _grafoPlanejamento );
	assert( NULL != _planoRelaxado );

	ultimoNivel = _grafoPlanejamento;
	while( NULL != ultimoNivel->proximo )
		ultimoNivel = ultimoNivel->proximo;

/*	copia no estado objetivo do nivel as proposicoes do estado objetivo */

	objetivoNivel = construirCopiaEstado( _problema->numProposicoes
			, _problema->objetivo );
	if( NULL == objetivoNivel )
		return( -1L );

/*	cria um estado "vazio" para os objetivos do nivel anterior */

	objetivoNivelAnterior = construirEstado( _problema->numProposicoes
				, _problema->proposicoes, "" );
	if( NULL == objetivoNivelAnterior )
		return( -1L );

/*	percorre o grafo de planejamento do ultimo nivel para o primeiro */

	while( ultimoNivel != _grafoPlanejamento )
		{
/*	percorre cada proposicao do objetivo do nivel */

		for( i = 0; _problema->numProposicoes > i; i++ )
			{
			if( INDEFINIDO == objetivoNivel[ i ].valorVerdade )
				continue;

			for( j = 0; ultimoNivel->numProposicoes > j; j++ )
				{
				if( i == ultimoNivel->verticeProposicao[ j ].numProposicao
						&& ultimoNivel->verticeProposicao[ j ].valorVerdade
						== objetivoNivel[ i ].valorVerdade )
					break;
				}

			assert( ultimoNivel->numProposicoes > j );

/*	procura acoes que adicionam a proposicao */

			arestaEscolhida = NULL;

			arestaEfeito = ultimoNivel->arestaEfeito;
			while( NULL != arestaEfeito )
				{
				if( ultimoNivel->verticeProposicao + j != arestaEfeito->vertice2 )
					{
					arestaEfeito = arestaEfeito->proxima;
					continue;
					}

/*	verifica se a acao ja foi adicionada a solucao */

//				acaoPlano = _planoRelaxado;
//				while( NULL != acaoPlano )
//					{
//					if( acaoPlano->acaoEscolhida == arestaEfeito->vertice1->numProposicao
//							&& acaoPlano->numeroNivel == ultimoNivel->numeroNivel )
//						{
//						arestaEscolhida = NULL;
//						break;
//						}
//
//					acaoPlano = acaoPlano->proxima;
//					}
				for( k = 0L; numeroAcoes > k; k++ )
					{
					if( _planoRelaxado[ k ].acaoEscolhida == arestaEfeito->vertice1->numProposicao
							&& _planoRelaxado[ k ].numeroNivel == ultimoNivel->numeroNivel )
						{
						arestaEscolhida = NULL;
						break;
						}
					}

				if( numeroAcoes > k )
					break;

/*	a escolha preferencial eh por acoes NOOP */

				if( 0 > arestaEfeito->vertice1->numProposicao )
					{
					arestaEscolhida = arestaEfeito;
					}
				else
					{
/*	se nao escolheu uma NOOP, escolhe a acao com menos pre-condicoes */

					if( NULL == arestaEscolhida )
						{
						arestaEscolhida = arestaEfeito;
						}
					else
						{
						if( 0 <= arestaEscolhida->vertice1->numProposicao
								&& arestaEfeito->vertice1->numeroPrecondicoes
								< arestaEscolhida->vertice1->numeroPrecondicoes )
							{
							arestaEscolhida = arestaEfeito;
							}
						}
					}

				arestaEfeito = arestaEfeito->proxima;
				}

			if( NULL != arestaEscolhida )
				{
/*	se nao eh NOOP, adiciona a acao escolhida na solucao */

				if( 0 <= arestaEscolhida->vertice1->numProposicao )
					{
//					novaEscolhaAcao = ( PEscolhaAcao )SM_MALLOC( "TEscolhaAcao"
//							, sizeof( TEscolhaAcao ) );
//					if( NULL == novaEscolhaAcao )
//						{
//						fprintf( stderr, "Erro na alocacao de memoria\n" );
//						return( -1L );
//						}
//
//					novaEscolhaAcao->numeroNivel = ultimoNivel->numeroNivel;
//					novaEscolhaAcao->acaoEscolhida = arestaEscolhida->vertice1->numProposicao;
//					if( NULL == planoSolucao )
//						novaEscolhaAcao->proxima = NULL;
//					else
//						novaEscolhaAcao->proxima = planoSolucao;
//					planoSolucao = novaEscolhaAcao;
					_planoRelaxado[ numeroAcoes ].numeroNivel
							= ultimoNivel->numeroNivel;
					_planoRelaxado[ numeroAcoes ].acaoEscolhida
							= arestaEscolhida->vertice1->numProposicao;
					_planoRelaxado[ numeroAcoes ].proxima = NULL;

					if( 0 < numeroAcoes )
						_planoRelaxado[ numeroAcoes - 1 ].proxima
								= _planoRelaxado + numeroAcoes;

					numeroAcoes++;
					}

/*	adiciona pre-condicoes da acao escolhida aos objetivos do nivel anterior */

				arestaPreCondicao = ultimoNivel->arestaPreCondicao;
				while( NULL != arestaPreCondicao )
					{
					if( arestaPreCondicao->vertice2 == arestaEscolhida->vertice1 )
						{
						objetivoNivelAnterior[ arestaPreCondicao->vertice1->numProposicao ].valorVerdade = arestaPreCondicao->vertice1->valorVerdade;
						}

					arestaPreCondicao = arestaPreCondicao->proxima;
					}
				}
			}

/*	passa para o nivel anterior do grafo */

		nivelAnterior = _grafoPlanejamento;
		while( ultimoNivel != nivelAnterior->proximo )
			nivelAnterior = nivelAnterior->proximo;

		ultimoNivel = nivelAnterior;

		copiarEstado( _problema->numProposicoes, objetivoNivel
				, objetivoNivelAnterior );

		for( i = 0; _problema->numProposicoes > i; i++ )
			objetivoNivelAnterior[ i ].valorVerdade = INDEFINIDO;
		}

	destruirEstado( objetivoNivel );
	destruirEstado( objetivoNivelAnterior );

	return( numeroAcoes );
	}

