/* /////////////////////////////////////////////////////////////////////////////
//	hillClimbing.c  -  18/10/2002 by aldeba
//
//	Implementacao do algoritmo de busca Hill-Climbing reforcado
///////////////////////////////////////////////////////////////////////////// */

/*	includes do fonte */

#include	<assert.h>
#include	<limits.h>
#include	<stdlib.h>
#include	<string.h>

#include	"graphplanRelaxado.h"
#include	"hillClimbing.h"
#include	"memoria.h"

/*
//	funcao heuristica a partir de um plano para uma tarefa relaxada
*/

long funcaoHeuristica( PEscolhaAcao _planoRelaxado )
	{
	PEscolhaAcao	acaoPlano;
	long int		numeroAcoes = 0L;

/*	conta o numero de acoes do plano para a tarefa relaxada */

	acaoPlano = _planoRelaxado;
	while( NULL != acaoPlano )
		{
		numeroAcoes++;

		acaoPlano = acaoPlano->proxima;
		}

	return( numeroAcoes );
	}

/*
//	retorna 1 se o operador eh uma acao util para o processo de busca em largura
//		Utilizada para a 'poda' da arvore de busca chamada 'Helpful Actions' 
*/

int ehAcaoUtil( PProblemaStrips _problema, POperadorStrips _operador
		, PEscolhaAcao _planoRelaxado )
	{
	PEscolhaAcao	acaoPlano;
	long int		i = 0L;

	assert( NULL != _problema );
	assert( NULL != _operador );

/*	verifica se operador adiciona efeitos das acoes do nivel 2 do plano */

	acaoPlano = _planoRelaxado;
	while( NULL != acaoPlano )
		{
		if( 2 != acaoPlano->numeroNivel )
			{
			acaoPlano = acaoPlano->proxima;
			continue;
			}

		for( i = 0L; _problema->numProposicoes > i; i++ )
			{
			if( INDEFINIDO != _problema->dominio[ acaoPlano->acaoEscolhida ].efeito[ i ].valorVerdade
					&& _problema->dominio[ acaoPlano->acaoEscolhida ].efeito[ i ].valorVerdade
					== _operador->efeito[ i ].valorVerdade )
				return( 1 );
			}

		acaoPlano = acaoPlano->proxima;
		}

	return( 0 );
	}

/*
//	retorna 1 se alguma acao do primeiro nivel do plano relaxado remove um
//		objetivo adicionado no atual estado da arvore de busca
//		Utilizada para a 'poda' da arvore de busca chamada 'Added Goal Deletion' 
*/

int acaoRemoveSubGoal( PProblemaStrips _problema, PArvoreBusca _arvoreBusca )
	{
	PEscolhaAcao	acaoPlano;
	long int		i = 0L;

	assert( NULL != _problema );
	assert( NULL != _arvoreBusca );

/*	procura proposicoes do objetivo adicionadas pela acao do noh da arvore */

	for( ; _problema->numProposicoes > i; i++ )
		{
		if( INDEFINIDO == _problema->objetivo[ i ].valorVerdade
				|| _problema->dominio[ _arvoreBusca->acao ].efeito[ i ].valorVerdade
				!= _problema->objetivo[ i ].valorVerdade )
			continue;

/*	verifica se alguma acao do nivel 2 do plano relaxado remove esse objetivo */

		acaoPlano = _arvoreBusca->planoRelaxado;
		while( NULL != acaoPlano )
			{
			if( 2 != acaoPlano->numeroNivel )
				{
				acaoPlano = acaoPlano->proxima;
				continue;
				}

			if( ( VERDADEIRO == _problema->dominio[ acaoPlano->acaoEscolhida ].efeito[ i ].valorVerdade
					&& FALSO == _problema->objetivo[ i ].valorVerdade )
					|| ( FALSO == _problema->dominio[ acaoPlano->acaoEscolhida ].efeito[ i ].valorVerdade
					&& VERDADEIRO == _problema->objetivo[ i ].valorVerdade ) )
				return( 1 );

			acaoPlano = acaoPlano->proxima;
			}
		}

	return( 0 );
	}

/* /////////////////////////////////////////////////////////////////////////////
// arvore de busca
///////////////////////////////////////////////////////////////////////////// */

/*
//	funcao para destruir uma arvore de busca em largura
*/

void destruirArvoreBusca( PArvoreBusca _arvoreBusca )
	{
	PListaDescendentes	descendente;
	PListaDescendentes	auxiliar;
	long int			i = 0L;

	assert( NULL != _arvoreBusca );

	if( NULL != _arvoreBusca->estadoVisitado )
		{
		for( i = 0; _arvoreBusca->numEstadosVisitados > i; i++ )
			destruirEstado( _arvoreBusca->estadoVisitado[ i ] );

		SM_FREE( _arvoreBusca->estadoVisitado );
		}

	if( NULL != _arvoreBusca->estado )
		destruirEstado( _arvoreBusca->estado );
	if( NULL != _arvoreBusca->grafoRelaxado )
		destruirGrafoPlanejamento( _arvoreBusca->grafoRelaxado );
	if( NULL != _arvoreBusca->planoRelaxado )
//		destruirEscolhaAcoes( _arvoreBusca->planoRelaxado );
		SM_FREE( _arvoreBusca->planoRelaxado );

	descendente = _arvoreBusca->descendente;
	while( NULL != descendente )
		{
		auxiliar = descendente->proximo;

		destruirArvoreBusca( descendente->nohFilho );
		SM_FREE( descendente );

		descendente = auxiliar;
		}

	SM_FREE( _arvoreBusca );
	}

/*
//	funcao que expande um nivel de profundidade da arvore de busca (largura)
*/

void expandeNivelArvoreBusca( PProblemaStrips _problema
		, long int _numEstadosVisitados, PEstado * _estadoVisitado
		, PArvoreBusca _arvoreBusca, int _criterio )
	{
	PEstado				estado;
	TProblemaStrips		subProblema;
	PArvoreBusca		raizArvore;
	PArvoreBusca		novoNoharvore;
	PListaDescendentes	novoDescendente;
	PListaDescendentes	descendente;
	long int			i = 0L;
	long int			j = 0L;

	assert( NULL != _problema );
	assert( 0 < _numEstadosVisitados );
	assert( NULL != _estadoVisitado );
	assert( NULL != _arvoreBusca );

/*	procura a raiz da arvore */

	raizArvore = _arvoreBusca;
	while( NULL != raizArvore->nohPai )
		raizArvore = raizArvore->nohPai;

/*	se o noh eh uma folha da arvore, expande um nivel */

	if( NULL == _arvoreBusca->descendente )
		{
		for( ; _problema->numOperadores > i; i++ )
			{
			if( 1 == ehAplicavel( _problema->numProposicoes
					, _arvoreBusca->estado, _problema->dominio + i )
					&& 1 == ehAcaoUtil( _problema, _problema->dominio + i
					, _arvoreBusca->planoRelaxado ) )
				{
				estado = estadoResultante( _problema->numProposicoes
						, _arvoreBusca->estado, _problema->dominio + i );

/*	verifica se esse estado ja foi visitado */

				for( j = 0; _numEstadosVisitados > j; j++ )
					{
					if( 0 == compararEstados( _problema->numProposicoes
							, estado, _estadoVisitado[ j ] ) )
						break;
					}

				if( _numEstadosVisitados > j )
					{
					destruirEstado( estado );
					continue;
					}

				for( j = 0; raizArvore->numEstadosVisitados > j; j++ )
					{
					if( 0 == compararEstados( _problema->numProposicoes
							, estado, raizArvore->estadoVisitado[ j ] ) )
						break;
					}

				if( raizArvore->numEstadosVisitados > j )
					{
					destruirEstado( estado );
					continue;
					}

#if		!defined( NDEBUG )

				fprintf( stderr, "Acao aplicavel: %s\n"
						, _problema->dominio[ i ].nome );
				escreverEstado( _problema->numProposicoes, _problema->proposicoes
						, estado, stderr );
#endif	/* !defined( NDEBUG ) */

/*	insere o novo estado na lista de estados visitados da arvore */

				if( NULL == raizArvore->estadoVisitado )
					raizArvore->estadoVisitado = ( PEstado * )SM_MALLOC( "PEstado"
							, ++raizArvore->numEstadosVisitados * sizeof( PEstado ) );
				else
					raizArvore->estadoVisitado = ( PEstado * )SM_REALLOC( "PEstado"
							, raizArvore->estadoVisitado
							, ++raizArvore->numEstadosVisitados * sizeof( PEstado ) );
				if( NULL == raizArvore->estadoVisitado )
					{
					fprintf( stderr, "Erro na alocacao de memoria\n" );
					return;
					}

				raizArvore->estadoVisitado[ raizArvore->numEstadosVisitados - 1 ] = construirCopiaEstado( _problema->numProposicoes
						, estado );
				if( NULL == raizArvore->estadoVisitado[ raizArvore->numEstadosVisitados - 1 ] )
					return;

/*	cria um novo noh para a arvore */

				novoNoharvore = ( PArvoreBusca )SM_MALLOC( "TArvoreBusca"
						, sizeof( TArvoreBusca ) );
				if( NULL == novoNoharvore )
					{
					fprintf( stderr, "Erro na alocacao de memoria\n" );
					return;
					}

				novoNoharvore->nohPai = _arvoreBusca;
				novoNoharvore->acao = i;
				novoNoharvore->numEstadosVisitados = 0L;
				novoNoharvore->estadoVisitado = NULL;
				novoNoharvore->estado = estado;
				novoNoharvore->grafoRelaxado = NULL;
				novoNoharvore->planoRelaxado = NULL;
				novoNoharvore->descendente = NULL;

/*	aloca memoria para o plano solucao, que nao deve ter mais acoes que */
/*		duas vezes o tamanho do dominio */

				novoNoharvore->planoRelaxado = ( PEscolhaAcao )SM_MALLOC( "TEscolhaAcao"
						, 2 * _problema->numOperadores * sizeof( TEscolhaAcao ) );
				if( NULL == novoNoharvore->planoRelaxado )
					{
					fprintf( stderr, "Erro na alocacao de memoria\n" );
					return;
					}

/*	insere o noh na arvore */

				novoDescendente = ( PListaDescendentes )SM_MALLOC( "TListaDescendentes"
						, sizeof( TListaDescendentes ) );
				if( NULL == novoDescendente )
					{
					fprintf( stderr, "Erro na alocacao de memoria\n" );
					return;
					}

				novoDescendente->nohFilho = novoNoharvore;
				if( NULL == _arvoreBusca->descendente )
					novoDescendente->proximo = NULL;
				else
					novoDescendente->proximo = _arvoreBusca->descendente;
				_arvoreBusca->descendente = novoDescendente;

/*	constroi um grafo de planejamento para a tarefa relaxada */
	
				subProblema.numProposicoes = _problema->numProposicoes;
				subProblema.proposicoes = _problema->proposicoes;
				subProblema.numOperadores = _problema->numOperadores;
				subProblema.dominio = _problema->dominio;
				subProblema.estadoInicial = estado;
				subProblema.objetivo = _problema->objetivo;

				novoNoharvore->grafoRelaxado = construirGrafoRelaxado( &subProblema );
				if( NULL == novoNoharvore->grafoRelaxado )
					{
					novoNoharvore->heuristica = LONG_MAX;
					destruirArvoreBusca( novoNoharvore );

					descendente = _arvoreBusca->descendente->proximo;
					SM_FREE( _arvoreBusca->descendente );
					_arvoreBusca->descendente = descendente;

					continue;
					}

/*	extrai uma solucao relaxada do grafo e retorna o numero de acoes da mesma */

				novoNoharvore->heuristica = extrairSolucaoRelaxada( &subProblema
						, novoNoharvore->grafoRelaxado
						, novoNoharvore->planoRelaxado );

//				novoNoharvore->heuristica = funcaoHeuristica( novoNoharvore->planoRelaxado );

				if( -1 == novoNoharvore->heuristica
						|| ( 0 != novoNoharvore->heuristica
						&& 0 != acaoRemoveSubGoal( _problema, novoNoharvore ) ) )
					{
					destruirArvoreBusca( novoNoharvore );

					descendente = _arvoreBusca->descendente->proximo;
					SM_FREE( _arvoreBusca->descendente );
					_arvoreBusca->descendente = descendente;

					continue;
					}

#if		!defined( NDEBUG )

				if( 0 != novoNoharvore->heuristica )
					{
					fprintf( stderr, "Plano relaxado: \n" );
					escreverEscolhaAcoes( &subProblema, novoNoharvore->planoRelaxado
							, stderr );
					}
				fprintf( stderr, "Heuristica = %ld\n", novoNoharvore->heuristica );
#endif	/* !defined( NDEBUG ) */

/*	depois de calcular a heuristica, pode liberar a memoria ocupada pelo grafo */

				destruirGrafoPlanejamento( novoNoharvore->grafoRelaxado );
				novoNoharvore->grafoRelaxado = NULL;

/*	se o criterio for velocidade, para assim que chegar a um estado melhor */

				if( CRITERION_SPEED ==  _criterio
						&& novoNoharvore->heuristica < raizArvore->heuristica )
					return;
				}
			}
		}
	else
		{
/*	se o noh nao eh uma folha da arvore, desce um nivel na profundidade */

		descendente = _arvoreBusca->descendente;
		while( NULL != descendente )
			{
#if		!defined( NDEBUG )

			fprintf( stderr, "Expandindo o ramo da acao: %s\n"
					, _problema->dominio[ descendente->nohFilho->acao ].nome );
#endif	/* !defined( NDEBUG ) */

			expandeNivelArvoreBusca( _problema, _numEstadosVisitados
					, _estadoVisitado, descendente->nohFilho, _criterio );

			descendente = descendente->proximo;
			}
		}
	}

/*
//	funcao que faz uma busca em largura na arvore pelo noh de menor heuristica
*/

PArvoreBusca buscaNohArvore( PArvoreBusca _arvoreBusca )
	{
	PArvoreBusca		nohArvoreBusca;
	PArvoreBusca		ramoArvoreBusca;
	PListaDescendentes	descendente;

	assert( NULL != _arvoreBusca );

/*	se o noh eh uma folha da arvore, o noh de melhor heuristica e ele mesmo */

	if( NULL == _arvoreBusca->descendente )
		return( _arvoreBusca );

/*	se o noh nao eh uma folha da arvore, desce um nivel na profundidade */

	nohArvoreBusca = _arvoreBusca;

	descendente = _arvoreBusca->descendente;
	while( NULL != descendente )
		{
		ramoArvoreBusca = buscaNohArvore( descendente->nohFilho );

		if( ramoArvoreBusca->heuristica < nohArvoreBusca->heuristica )
			nohArvoreBusca = ramoArvoreBusca;

		descendente = descendente->proximo;
		}

	return( nohArvoreBusca );
	}

/* /////////////////////////////////////////////////////////////////////////////
// lista ordenada de acoes
///////////////////////////////////////////////////////////////////////////// */

/*
//	funcao para destruir uma lista ordenada de acoes
*/

void destruirListaOrdenadaAcoes( PListaOrdenadaAcoes _lista )
	{
	PListaOrdenadaAcoes	acao;
	PListaOrdenadaAcoes	auxiliar;

	assert( NULL != _lista );

	acao = _lista;
	while( NULL != acao )
		{
		auxiliar = acao->proxima;

		SM_FREE( acao );

		acao = auxiliar;
		}
	}

/*
//	funcao para escrever uma lista ordenada de acoes em um stream
*/

void escreverListaOrdenadaAcoes( PProblemaStrips _problema
		, PListaOrdenadaAcoes _lista, FILE * _stream )
	{
	PListaOrdenadaAcoes	acao;
	long int			i = 1L;

	assert( NULL != _problema );
	assert( NULL != _lista );
	assert( NULL != _stream );

	acao = _lista;
	while( NULL != acao )
		{
		fprintf( _stream, "%ld- %s\n", i++
				, _problema->dominio[ acao->acao ].nome );

		acao = acao->proxima;
		}
	}

/*
//	implementacao do algoritmo de busca Hill-Climbing reforcado
*/

PListaOrdenadaAcoes hillClimbingReforcado( PProblemaStrips _problema
		, int _criterio )
	{
	PListaOrdenadaAcoes	planoSolucao = NULL;
	PEstado *			estadoVisitado;
	PEstado				estado = NULL;
	TProblemaStrips		subProblema;
	PNivelGrafo			grafoRelaxado;
	PEscolhaAcao		planoRelaxado;
	PArvoreBusca		arvoreBuscaLargura = NULL;
	PArvoreBusca		nohMenorHeuristica;
	PArvoreBusca		nohPesquisa;
	long				heuristicaEstadoAtual;
	long int			numAcoesPlanoSolucao = 0L;
	long int			numEstadosVisitados = 0L;
	int					profundidadeArvore;
	const int			profundidadeCorte = 15;
	long int			i = 0L;

	assert( NULL != _problema );
	assert( CRITERION_SPEED ==  _criterio
			|| CRITERION_QUALITY == _criterio );

	SM_IMPRIME_OCUPACAO;

/*	aloca memoria para o plano solucao, que nao deve ter mais acoes que */
/*		duas vezes o tamanho do dominio */

	planoSolucao = ( PListaOrdenadaAcoes )SM_MALLOC( "TListaOrdenadaAcoes"
			, 2 * _problema->numOperadores * sizeof( TListaOrdenadaAcoes ) );
	if( NULL == planoSolucao )
		{
		fprintf( stderr, "Erro na alocacao de memoria\n" );
		return( NULL );
		}

/*	coloca o estado inicial no vetor de estados visitados */

	estadoVisitado = ( PEstado * )SM_MALLOC( "PEstado"
			,  2 * _problema->numOperadores * sizeof( PEstado ) );
	if( NULL == estadoVisitado )
		{
		fprintf( stderr, "Erro na alocacao de memoria\n" );
		return( NULL );
		}

	for( i = 0; 2 * _problema->numOperadores > i; i++ )
		{
		estadoVisitado[ i ] = construirCopiaEstado( _problema->numProposicoes
				, _problema->estadoInicial );
		if( NULL == estadoVisitado[ i ] )
			return( NULL );
		}

	numEstadosVisitados = 1L;

/*	copia no estado atual as proposicoes do estado inicial */

	estado = construirCopiaEstado( _problema->numProposicoes
			, _problema->estadoInicial );
	if( NULL == estado )
		return( NULL );

/*	enquanto nao atingiu o estado objetivo, realiza uma busca local */

	for( ;; )
		{
/*	aloca memoria para o plano solucao, que nao deve ter mais acoes que */
/*		duas vezes o tamanho do dominio */

		planoRelaxado = ( PEscolhaAcao )SM_MALLOC( "TEscolhaAcao"
				, 2 * _problema->numOperadores * sizeof( TEscolhaAcao ) );
		if( NULL == planoRelaxado )
			{
			fprintf( stderr, "Erro na alocacao de memoria\n" );
			return( NULL );
			}

/*	calcula a funcao heuristica para o estado atual */
	
		subProblema.numProposicoes = _problema->numProposicoes;
		subProblema.proposicoes = _problema->proposicoes;
		subProblema.numOperadores = _problema->numOperadores;
		subProblema.dominio = _problema->dominio;
		subProblema.estadoInicial = estado;
		subProblema.objetivo = _problema->objetivo;

		grafoRelaxado = construirGrafoRelaxado( &subProblema );
		if( NULL == grafoRelaxado )
			{
			heuristicaEstadoAtual = LONG_MAX;
			return( NULL );
			}

/*	extrai uma solucao relaxada do grafo e retorna o numero de acoes da mesma */

//		planoRelaxado = extrairSolucaoRelaxada( &subProblema, grafoRelaxado );
//		heuristicaEstadoAtual = funcaoHeuristica( planoRelaxado );
		heuristicaEstadoAtual = extrairSolucaoRelaxada( &subProblema
				, grafoRelaxado, planoRelaxado );

		if( -1 == heuristicaEstadoAtual )
			return( NULL );

		if( 0 == heuristicaEstadoAtual )
			break;

#if		!defined( NDEBUG )

		fprintf( stderr, "Estado atual: \n" );
		escreverEstado( _problema->numProposicoes, _problema->proposicoes
				, estado, stderr );

		fprintf( stderr, "Plano relaxado: \n" );
		escreverEscolhaAcoes( _problema, planoRelaxado, stderr );
		fprintf( stderr, "Heuristica = %ld\n", heuristicaEstadoAtual );
#endif	/* !defined( NDEBUG ) */

/*	efetua uma busca em largura ate atingir um estado de menor heuristica */

		arvoreBuscaLargura = ( PArvoreBusca )SM_MALLOC( "TArvoreBusca"
				, sizeof( TArvoreBusca ) );
		if( NULL == arvoreBuscaLargura )
			{
			fprintf( stderr, "Erro na alocacao de memoria\n" );
			return( NULL );
			}

		arvoreBuscaLargura->nohPai = NULL;
		arvoreBuscaLargura->acao = -1;
		arvoreBuscaLargura->numEstadosVisitados = 0L;
		arvoreBuscaLargura->estadoVisitado = NULL;

		arvoreBuscaLargura->estado = construirCopiaEstado( _problema->numProposicoes
				, estado );
		if( NULL == arvoreBuscaLargura->estado )
			return NULL;

		arvoreBuscaLargura->grafoRelaxado = grafoRelaxado;
		arvoreBuscaLargura->planoRelaxado = planoRelaxado;
		arvoreBuscaLargura->heuristica = heuristicaEstadoAtual;
		arvoreBuscaLargura->descendente = NULL;

		profundidadeArvore = 0;

		for( ;; )
			{
#if		!defined( NDEBUG )

			if( 0 < profundidadeArvore )
				fprintf( stderr, "Descendo para a profundidade %d\n"
						, profundidadeArvore );
#endif	/* !defined( NDEBUG ) */

			if( profundidadeCorte <= profundidadeArvore )
				{
#if		!defined( NDEBUG )

				fprintf( stderr, "Atingiu a profundidade de corte\n" );
#endif	/* !defined( NDEBUG ) */

				destruirEstado( estado );
				destruirArvoreBusca( arvoreBuscaLargura );
				if( NULL != planoSolucao )
//					destruirListaOrdenadaAcoes( planoSolucao );
					SM_FREE( planoSolucao );

				return( NULL );
				}

			expandeNivelArvoreBusca( _problema, numEstadosVisitados
					, estadoVisitado, arvoreBuscaLargura, _criterio );
			profundidadeArvore++;

			SM_IMPRIME_OCUPACAO;

			nohMenorHeuristica = buscaNohArvore( arvoreBuscaLargura );
			if( nohMenorHeuristica->heuristica < heuristicaEstadoAtual )
				break;
			}

/*	atualiza o estado atual e sua heuristica */

		copiarEstado( _problema->numProposicoes, estado
				, nohMenorHeuristica->estado );

		heuristicaEstadoAtual = nohMenorHeuristica->heuristica;

/*	concatena ao plano a sequencia das acoes necessarias */
/*		para atingir o estado de menor heuristica */

		nohPesquisa = nohMenorHeuristica;
		while( NULL != nohPesquisa->nohPai )
			{
			numAcoesPlanoSolucao++;

			nohPesquisa = nohPesquisa->nohPai;
			}

		for( i = 1; ; i++ )
			{
			if( NULL == nohMenorHeuristica->nohPai )
				break;

			planoSolucao[ numAcoesPlanoSolucao - i ].acao
					= nohMenorHeuristica->acao;
			if( 1 == i )
				planoSolucao[ numAcoesPlanoSolucao - i ].proxima = NULL;

			if( numAcoesPlanoSolucao - i - 1 >= 0 )
				planoSolucao[ numAcoesPlanoSolucao - i - 1 ].proxima
						= planoSolucao + numAcoesPlanoSolucao - i;

			nohMenorHeuristica = nohMenorHeuristica->nohPai;
			}

/*	adiciona os estados da arvore de busca ao vetor de estados visitados */

		nohPesquisa = nohMenorHeuristica;
		while( NULL != nohPesquisa->nohPai )
			{
			copiarEstado( _problema->numProposicoes
					, estadoVisitado[ numEstadosVisitados++ ]
					, nohPesquisa->estado );

			nohPesquisa = nohPesquisa->nohPai;
			}

		destruirArvoreBusca( arvoreBuscaLargura );

		SM_IMPRIME_OCUPACAO;
		}

	for( i = 0; 2 * _problema->numOperadores > i; i++ )
		destruirEstado( estadoVisitado[ i ] );

	SM_FREE( estadoVisitado );

	if( NULL != estado )
		destruirEstado( estado );
	if( NULL != grafoRelaxado )
		destruirGrafoPlanejamento( grafoRelaxado );
	if( NULL != planoRelaxado )
//		destruirEscolhaAcoes( planoRelaxado );
		SM_FREE( planoRelaxado );

	SM_IMPRIME_OCUPACAO;

	return( planoSolucao );
	}

