/* /////////////////////////////////////////////////////////////////////////////
//	hillClimbing.h  -  18/10/2002 by aldeba
//
//	Prototipos das funcoes do algoritmo de busca Hill-Climbing reforcado
///////////////////////////////////////////////////////////////////////////// */

/*	protecao contra referencias circulares */

#if			!defined( __HILLCLIMBING_H )

#define		__HILLCLIMBING_H

/*	includes do fonte */

#include	"graphplan.h"
#include	"strips.h"

/* /////////////////////////////////////////////////////////////////////////////
//	declaracoes de estruturas de dados
///////////////////////////////////////////////////////////////////////////// */

#define	CRITERION_SPEED		1
#define CRITERION_QUALITY	2

/*	arvore de busca */

struct arvoreBusca;

struct listaDescendentes {
	struct arvoreBusca *		nohFilho;
	struct listaDescendentes *	proximo;
};

typedef struct listaDescendentes	TListaDescendentes;
typedef TListaDescendentes *		PListaDescendentes;

struct arvoreBusca {
	struct arvoreBusca *		nohPai;
	long int					numEstadosVisitados;
	PEstado *					estadoVisitado;
	long int					acao;
	PEstado						estado;
	PNivelGrafo					grafoRelaxado;
	PEscolhaAcao				planoRelaxado;
	long						heuristica;
	PListaDescendentes			descendente;
};

typedef struct arvoreBusca			TArvoreBusca;
typedef TArvoreBusca *				PArvoreBusca;

/*	lista ordenada de acoes */

struct listaOrdenadaAcoes {
	long int					acao;
	struct listaOrdenadaAcoes *	proxima;
};

typedef struct listaOrdenadaAcoes	TListaOrdenadaAcoes;
typedef TListaOrdenadaAcoes *		PListaOrdenadaAcoes;

/* /////////////////////////////////////////////////////////////////////////////
// declaracoes de prototipos
///////////////////////////////////////////////////////////////////////////// */

long				funcaoHeuristica			( PEscolhaAcao _planoRelaxado );
int					ehAcaoUtil					( PProblemaStrips _problema, POperadorStrips _operador, PEscolhaAcao _planoRelaxado );
int					acaoRemoveSubGoal			( PProblemaStrips _problema, PArvoreBusca _arvoreBusca );

void				destruirArvoreBusca			( PArvoreBusca _arvoreBusca );
void				expandeNivelArvoreBusca		( PProblemaStrips _problema, long int _numEstadosVisitados, PEstado * _estadoVisitado, PArvoreBusca _arvoreBusca, int _criterio );
PArvoreBusca		buscaNohArvore				( PArvoreBusca _arvoreBusca );

void				destruirListaOrdenadaAcoes	( PListaOrdenadaAcoes _lista );
void				escreverListaOrdenadaAcoes	( PProblemaStrips _problema, PListaOrdenadaAcoes _lista, FILE * _stream );
PListaOrdenadaAcoes	hillClimbingReforcado		( PProblemaStrips _problema, int _criterio );
#endif		/* !defined( __HILLCLIMBING_H ) */

