/* /////////////////////////////////////////////////////////////////////////////
// memoria.c  -  18/11/2002 by aldeba
//
// Subsistema de alocacao de memoria
///////////////////////////////////////////////////////////////////////////// */

/*	includes do fonte */

#include	<assert.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>

#include	"memoria.h"

#if			!defined( NDEBUG )

/* /////////////////////////////////////////////////////////////////////////////
// subsistema de alocacao de memoria
///////////////////////////////////////////////////////////////////////////// */

PListaBlocosMemoria	subSistemaMemoria = NULL;

/* /////////////////////////////////////////////////////////////////////////////
// lista de blocos de memoria
///////////////////////////////////////////////////////////////////////////// */

/*
//	funcao para destruir uma lista ligada de blocos de memoria
*/

void destruirListaBlocosMemoria( PListaBlocosMemoria _lista )
	{
	PListaBlocosMemoria	bloco;
	PListaBlocosMemoria	auxiliar;

	assert( NULL != _lista );

	bloco = _lista;
	while( NULL != bloco )
		{
		auxiliar = bloco->proximo;

		if( NULL != bloco->ponteiro )
			free( bloco->ponteiro );

		free( bloco );

		bloco = auxiliar;
		}
	}

/*
//	funcao para escrever uma lista ligada de blocos de memoria em um stream
*/

void escreverListaBlocosMemoria( PListaBlocosMemoria _lista, FILE * _stream )
	{
	PListaBlocosMemoria	bloco;

	assert( NULL != _lista );
	assert( NULL != _stream );

	bloco = _lista;
	while( NULL != bloco )
		{
		if( NULL != bloco->ponteiro )
			fprintf( _stream, "Bloco de %s - %ld bytes alocado em %s - %u\n"
					, bloco->tipoDado, bloco->tamanho
					, bloco->programa, bloco->numeroLinha );

		bloco = bloco->proximo;
		}
	}

/*
//	funcao para escrever a ocupacao de um subsistema de memoria em um stream
*/

void escreverOcupacaoMemoria( PListaBlocosMemoria _lista, FILE * _stream )
	{
	PListaBlocosMemoria	bloco;
	long int			totalMemoria = 0L;

	assert( NULL != _lista );
	assert( NULL != _stream );

	bloco = _lista;
	while( NULL != bloco )
		{
		if( NULL != bloco->ponteiro )
			totalMemoria += bloco->tamanho;

		bloco = bloco->proximo;
		}

	fprintf( _stream, "%ld bytes alocados pelo subsistema de memoria\n"
			, totalMemoria );
	}

/*
//	funcao para adicionar um bloco a uma lista ligada de blocos de memoria
*/

void * adicionaBlocoMemoria( PListaBlocosMemoria * _lista, char * _programa
		, unsigned int _numeroLinha, char * _tipoDado, long _tamanho )
	{
	PListaBlocosMemoria	novoBloco;
	PListaBlocosMemoria	ultimoBloco;

	assert( NULL != _lista );
	assert( NULL != _tipoDado );
	assert( 0 < _tamanho );

	ultimoBloco = *_lista;
	while( NULL != ultimoBloco
			&& NULL != ultimoBloco->proximo )
		{
		ultimoBloco = ultimoBloco->proximo;
		}

	novoBloco = ( PListaBlocosMemoria )malloc( sizeof( TListaBlocosMemoria ) );
	if( NULL == novoBloco )
		{
		fprintf( stderr, "Erro na alocacao de memoria\n" );
		return( NULL );
		}

	strcpy( novoBloco->programa, _programa );
	novoBloco->numeroLinha = _numeroLinha;
	strcpy( novoBloco->tipoDado, _tipoDado );

	novoBloco->ponteiro = malloc( _tamanho );
	if( NULL == novoBloco->ponteiro )
		{
		fprintf( stderr, "Erro na alocacao de memoria\n" );
		return( NULL );
		}

	novoBloco->tamanho = _tamanho;
	novoBloco->proximo = NULL;

	if( NULL == ultimoBloco )
		*_lista = novoBloco;
	else
		ultimoBloco->proximo = novoBloco;

#if			defined( SM_TRACE )

	fprintf( stderr, "Novo bloco de memoria: %s - %ld\n", _tipoDado, _tamanho );
#endif		/* defined( SM_TRACE ) */

	return( novoBloco->ponteiro );
	}

/*
//	funcao para redimensionar um bloco de uma lista ligada de blocos de memoria
*/

void * redimensionaBlocoMemoria( PListaBlocosMemoria * _lista, char * _programa
		, unsigned int _numeroLinha, char * _tipoDado, void * _ponteiro
		, long _tamanho )
	{
	PListaBlocosMemoria	pesquisa;

	assert( NULL != _lista );
	assert( NULL != _tipoDado );
	assert( NULL != _ponteiro );
	assert( 0 < _tamanho );

	pesquisa = *_lista;
	while( NULL != pesquisa )
		{
		if( 0 == pesquisa->ponteiro - _ponteiro )
			{
			if( 0 != strcmp( pesquisa->tipoDado, _tipoDado ) )
				{
				fprintf( stderr, "Tentativa de redimensionar bloco de tipo incorreto em %s - %u\n"
						, _programa, _numeroLinha );

				return( NULL );
				}

			pesquisa->ponteiro = realloc( _ponteiro, _tamanho );
			if( NULL == pesquisa->ponteiro )
				{
				fprintf( stderr, "Erro na realocacao de memoria\n" );
				return( NULL );
				}

#if			defined( SM_TRACE )

			fprintf( stderr, "Bloco de memoria redimensionado: %s - %ld -> %ld\n"
					, pesquisa->tipoDado, pesquisa->tamanho, _tamanho );
#endif		/* defined( SM_TRACE ) */

			pesquisa->tamanho = _tamanho;

			return( pesquisa->ponteiro );
			}

		pesquisa = pesquisa->proximo;
		}

	fprintf( stderr, "Tentativa de redimensionar bloco invalido em %s - %u\n"
			, _programa, _numeroLinha );

	return( NULL );
	}

/*
//	funcao para remover um bloco de uma lista ligada de blocos de memoria
*/

void removeBlocoMemoria( PListaBlocosMemoria * _lista, char * _programa
		, unsigned int _numeroLinha, void * _ponteiro )
	{
	PListaBlocosMemoria	pesquisa;
	PListaBlocosMemoria	anterior = NULL;

	assert( NULL != _lista );
	assert( NULL != _ponteiro );

	pesquisa = *_lista;
	while( NULL != pesquisa )
		{
		if( 0 == pesquisa->ponteiro - _ponteiro )
			{
			if( NULL != anterior )
				anterior->proximo = pesquisa->proximo;
			else
				*_lista = pesquisa->proximo;

#if			defined( SM_TRACE )

			fprintf( stderr, "Bloco de memoria liberado: %s - %ld\n"
					, pesquisa->tipoDado, pesquisa->tamanho );
#endif		/* defined( SM_TRACE ) */

			free( pesquisa->ponteiro );
			free( pesquisa );

			return;
			}

		anterior = pesquisa;
		pesquisa = pesquisa->proximo;
		}

	fprintf( stderr, "Tentativa de liberar bloco invalido em %s - %u\n"
			, _programa, _numeroLinha );
	}
#endif		/* !defined( NDEBUG ) */

