/* /////////////////////////////////////////////////////////////////////////////
// memoria.h  -  18/11/2002 by aldeba
//
// Estruturas de dados e prototipos das funcoes
//    do subsistema de alocacao de memoria
///////////////////////////////////////////////////////////////////////////// */

/*	protecao contra referencias circulares */

#if			!defined( __MEMORIA_H )

#define		__MEMORIA_H

/*	includes do fonte */

#include	<stdio.h>

/* /////////////////////////////////////////////////////////////////////////////
// declaracoes de estruturas de dados
///////////////////////////////////////////////////////////////////////////// */

// lista de blocos de memoria

struct listaBlocosMemoria {
	char						programa[ 51 ];
	unsigned int				numeroLinha;
	char						tipoDado[ 51 ];
	void *						ponteiro;
	long						tamanho;
	struct listaBlocosMemoria *	proximo;
};

typedef struct listaBlocosMemoria	TListaBlocosMemoria;
typedef TListaBlocosMemoria *		PListaBlocosMemoria;

// objeto do subsistema de alocacao de memoria

extern PListaBlocosMemoria	subSistemaMemoria;

/* /////////////////////////////////////////////////////////////////////////////
// declaracoes de prototipos
///////////////////////////////////////////////////////////////////////////// */

void	destruirListaBlocosMemoria	( PListaBlocosMemoria _lista );
void	escreverListaBlocosMemoria	( PListaBlocosMemoria _lista, FILE * _stream );
void	escreverOcupacaoMemoria		( PListaBlocosMemoria _lista, FILE * _stream );
void *	adicionaBlocoMemoria		( PListaBlocosMemoria * _lista, char * _programa, unsigned int _numeroLinha, char * _tipoDado, long _tamanho );
void *	redimensionaBlocoMemoria	( PListaBlocosMemoria * _lista, char * _programa, unsigned int _numeroLinha, char * _tipoDado, void * _ponteiro, long _tamanho );
void	removeBlocoMemoria			( PListaBlocosMemoria * _lista, char * _programa, unsigned int _numeroLinha, void * _ponteiro );

// macros para utilizacao do subsistema de memoria

#if			!defined( NDEBUG )

#define	SM_INICIALIZA \
	( subSistemaMemoria = NULL )

#define	SM_FINALIZA \
	( void )( NULL == subSistemaMemoria \
		|| ( fprintf( stderr, "\nBlocos de memoria nao liberados\n\n" ) \
		, escreverListaBlocosMemoria( subSistemaMemoria, stderr ), 0 ) )

#define	SM_IMPRIME_OCUPACAO \
	( escreverOcupacaoMemoria( subSistemaMemoria, stderr ) )

#define	SM_MALLOC(_tipoDado,_tamanho) \
	( adicionaBlocoMemoria( &subSistemaMemoria, __FILE__, __LINE__ \
			, _tipoDado, _tamanho ) )

#define	SM_REALLOC(_tipoDado,_ponteiro,_tamanho) \
	( redimensionaBlocoMemoria( &subSistemaMemoria, __FILE__, __LINE__ \
			, _tipoDado, _ponteiro, _tamanho ) )

#define	SM_FREE(_ponteiro) \
	( removeBlocoMemoria( &subSistemaMemoria, __FILE__, __LINE__, _ponteiro ) )
#else		/* !defined( NDEBUG ) */

#define	SM_INICIALIZA \
	( ( void )( 0 ) )

#define	SM_FINALIZA \
	( ( void )( 0 ) )

#define	SM_IMPRIME_OCUPACAO \
	( ( void )( 0 ) )

#define	SM_MALLOC(_tipoDado,_tamanho) \
	( malloc( _tamanho ) )

#define	SM_REALLOC(_tipoDado,_ponteiro,_tamanho) \
	( realloc( _ponteiro, _tamanho ) )

#define	SM_FREE(_ponteiro) \
	( free( _ponteiro ) )
#endif		/* !defined( NDEBUG ) */
#endif		/* !defined( __MEMORIA_H ) */

