/* /////////////////////////////////////////////////////////////////////////////
// principal.c  -  17/04/2002 by aldeba
//
// Ponto de entrada do programa FF
///////////////////////////////////////////////////////////////////////////// */

/*	includes do fonte */

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>

#include	"strips.h"
#include	"hillClimbing.h"
#include	"memoria.h"

#include	"graphplanRelaxado.h"

#define	CHAVE_AJUDA			"--help"
#define	CHAVE_PROBLEMA		"--problem="
#define	CHAVE_CRITERIO		"--criterion="

/*
//	ponto de entrada do programa
*/

int main( int _argc, char * _argv[] )
	{
	char *				nomeArquivoProblema = NULL;
	FILE *				arquivoProblema = NULL;
	TProblemaStrips		problema;
	PListaOrdenadaAcoes	planoSolucao;
	int					criterio = CRITERION_SPEED;
	int					i;
	int					retorno = 0;

	SM_INICIALIZA;
	
	fprintf( stdout, "FF  -  Versao 1.0 por Aldebaran Perseke"
		" em 18/10/2002\n" );
	fprintf( stdout, "\n" );

	construirProblemaStrips( &problema );

/*	obtem os argumentos da linha de comandos */

	for( i = 1; i < _argc; i++ )
		{
		if( strncmp( _argv[ i ], CHAVE_AJUDA, sizeof( CHAVE_AJUDA ) - 1 ) == 0 )
			{
			fprintf( stdout, "uso: %s [%sarquivo-problema] [%s{SPEED|QUALITY]%s\n\t"
				, _argv[ 0 ], CHAVE_PROBLEMA, CHAVE_CRITERIO, CHAVE_AJUDA );
			return( 0 );
			}

		if( strncmp( _argv[ i ], CHAVE_PROBLEMA, sizeof( CHAVE_PROBLEMA ) - 1 )
				== 0 )
			{
			nomeArquivoProblema = _argv[ i ] + sizeof( CHAVE_PROBLEMA ) - 1;
			continue;
			}

		if( strncmp( _argv[ i ], CHAVE_CRITERIO, sizeof( CHAVE_CRITERIO ) - 1 )
				== 0 )
			{
			if( strcmp( _argv[ i ] + sizeof( CHAVE_CRITERIO ) - 1, "SPEED" )
					== 0 )
				{
				criterio = CRITERION_SPEED;
				continue;
				}

			if( strcmp( _argv[ i ] + sizeof( CHAVE_CRITERIO ) - 1, "QUALITY" )
					== 0 )
				{
				criterio = CRITERION_QUALITY;
				continue;
				}

			fprintf( stdout, "%s: opcao invalida %s para a chave %s\n"
				, _argv[ 0 ], _argv[ i ] + sizeof( CHAVE_CRITERIO ) - 1
				, CHAVE_CRITERIO );
			return( 0 );
			}

		fprintf( stdout, "%s: argumento invalido %s\n"
			, _argv[ 0 ], _argv[ i ] );
		return( 0 );
		}

/*	abre o arquivo com o problema */

	if( NULL == nomeArquivoProblema )
		{
		arquivoProblema = stdin;
		}
	else
		{
		arquivoProblema = fopen( nomeArquivoProblema, "rt" );
		if( NULL == arquivoProblema )
			{
			fprintf( stderr, "Erro na abertura do arquivo %s\n"
				, nomeArquivoProblema );
			return( -1 );
			}
		}

/*	le o arquivo com o problema */

	if( NULL != nomeArquivoProblema )
		fprintf( stdout, "arquivo com o problema: %s\n", nomeArquivoProblema );

	if( lerProblemaSTRIPS( arquivoProblema, &problema ) != 0 )
		return( -1 );

	fprintf( stdout, "problema lido\n" );

/*	fecha o arquivo com o problema */

	if( NULL != nomeArquivoProblema )
		fclose( arquivoProblema );

#if		!defined( NDEBUG )

	escreverProblemaStrips( &problema, stderr );
#endif	/* !defined( NDEBUG ) */

/*	resolve o problema de planejamento e imprime a solucao */

	planoSolucao = hillClimbingReforcado( &problema, criterio );
	if( NULL == planoSolucao )
		{
		fprintf( stdout, "Nao foi possivel encontrar uma solucao\n" );
		}
	else
		{
		fprintf( stdout, "Plano solucao para o problema:\n" );
		escreverListaOrdenadaAcoes( &problema, planoSolucao, stdout );
		}
	
/*	libera memoria ocupada pelo problema e pela solucao */

	if( NULL != planoSolucao )
//		destruirListaOrdenadaAcoes( planoSolucao );
		SM_FREE( planoSolucao );

//	{
//	PNivelGrafo		grafoRelaxado;
//	PEscolhaAcao	planoRelaxado;
//	long			heuristicaEstadoAtual;
//
//	grafoRelaxado = construirGrafoRelaxado( &problema );
//	if( NULL != grafoRelaxado )
//		{
//		planoRelaxado = extrairSolucaoRelaxada( &problema, grafoRelaxado );
//		heuristicaEstadoAtual = funcaoHeuristica( planoRelaxado );
//		fprintf( stderr, "Heuristica = %ld\n", heuristicaEstadoAtual );
//		}
//	}

	destruirProblemaStrips( &problema );

	SM_FINALIZA;

	return( retorno );
	}

