/* /////////////////////////////////////////////////////////////////////////////
// strips.c  -  19/05/2002 by aldeba
//
// Funcoes para a descricao de problemas de planejamento STRIPS
///////////////////////////////////////////////////////////////////////////// */

/*	includes do fonte */

#include	<assert.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>

#include	"strips.h"
#include	"memoria.h"

/* /////////////////////////////////////////////////////////////////////////////
// proposicao
///////////////////////////////////////////////////////////////////////////// */

/*
//	funcao para destruir um vetor de proposicoes
*/

void destruirProposicoes( PProposicao _vetor )
	{
	assert( NULL != _vetor );

	SM_FREE( _vetor );
	}

/*
//	funcao para comparar duas proposicoes
*/

int compararProposicoes( const void * _p1, const void * _p2 )
	{
	return( strcmp( ( ( PProposicao )_p1 )->nome
			, ( ( PProposicao )_p2 )->nome ) );
	}

/*
//	funcao para escrever um vetor de proposicoes em um stream
*/

void escreverProposicoes( long int _numProposicoes, PProposicao _vetor
		, FILE * _stream )
	{
	long int	i = 0L;

	assert( 0 < _numProposicoes );
	assert( NULL != _vetor );
	assert( NULL != _stream );

	for( ; _numProposicoes > i; i++ )
		fprintf( _stream, "%s\n", _vetor[ i ].nome );
	}

/*
//	funcao para inserir uma proposicao em um vetor de proposicoes
*/

PProposicao inserirProposicao( long int _numProposicoes, PProposicao _vetor
		, char * _proposicao )
	{
	PProposicao	novoVetor;

	assert( NULL != _proposicao );

	if( NULL == _vetor )
		{
		assert( 1 == _numProposicoes );

		novoVetor = ( PProposicao )SM_MALLOC( "TProposicao"
				, _numProposicoes * sizeof( TProposicao ) );
		if( NULL == novoVetor )
			{
			fprintf( stderr, "Erro na alocacao de memoria\n" );
			return( NULL );
			}
		}
	else
		{
		assert( 1 < _numProposicoes );

		novoVetor = ( PProposicao )SM_REALLOC( "TProposicao", _vetor
				, _numProposicoes * sizeof( TProposicao ) );
		if( NULL == novoVetor )
			{
			fprintf( stderr, "Erro na alocacao de memoria\n" );
			return( NULL );
			}
		}

	strcpy( novoVetor[ _numProposicoes - 1 ].nome, _proposicao );
	novoVetor[ _numProposicoes - 1 ].numProposicao = _numProposicoes;

/*	ordena o vetor apos a insercao */

	if( 1 < _numProposicoes )
		qsort( novoVetor, _numProposicoes, sizeof( TProposicao )
				, compararProposicoes );

	return( novoVetor );
	}

/*
//	funcao para retornar o indice de uma proposicao em um vetor de proposicoes
*/

long int procurarProposicao( long int _numProposicoes, PProposicao _vetor
		, char * _proposicao )
	{
	PProposicao	elemento;
	
	elemento = bsearch( _proposicao, _vetor, _numProposicoes
			, sizeof( TProposicao ), compararProposicoes );
	if( NULL == elemento )
		return( -1 );

	return( elemento - _vetor );
	}

/* /////////////////////////////////////////////////////////////////////////////
// estado
///////////////////////////////////////////////////////////////////////////// */

/*
//	funcao para construir um vetor de estado dadas as proposicoes de uma linha
*/

PEstado construirEstado( long int _numProposicoes, PProposicao _vetor
		, char * _linha )
	{
	PEstado		novoEstado;
	char		proposicao[ 101 ];
	const char	delimitador = ';';
	char *		p = _linha;
	char *		q = _linha;
	long int	i = 0L;

	assert( 0 < _numProposicoes );
	assert( NULL != _vetor );
	assert( NULL != _linha );

/*	cria o estado com todos os valores das proposicoes indefinidos */

	novoEstado = ( PEstado )SM_MALLOC( "TEstado"
			, _numProposicoes * sizeof( TEstado ) );
	if( NULL == novoEstado )
		{
		fprintf( stderr, "Erro na alocacao de memoria\n" );
		return( NULL );
		}

	for( ; _numProposicoes > i; i++ )
		novoEstado[ i ].valorVerdade = INDEFINIDO;

/*	liga/desliga os valores das proposicoes contidas na linha */

	for( ;; )
		{
		if( '\0' == *q
				&& 0 == q - p )
			break;

		if( delimitador == *q
				|| '\0' == *q )
			{
			strncpy( proposicao, p, q - p );
			proposicao[ q - p ] = '\0';

			if( '~' != proposicao[ 0 ] )
				{
				i = procurarProposicao( _numProposicoes, _vetor, proposicao );

				assert( -1 != i );

				novoEstado[ i ].valorVerdade = VERDADEIRO;
				}
			else
				{
				i = procurarProposicao( _numProposicoes, _vetor, proposicao + 1 );

				assert( -1 != i );

				novoEstado[ i ].valorVerdade = FALSO;
				}

			if( '\0' == *q )
				break;

			p = q + 1;
			}

		q++;
		}

	return( novoEstado );
	}

/*
//	funcao para contruir uma copia de um vetor de estado
*/

PEstado construirCopiaEstado( long int _numProposicoes, PEstado _origem )
	{
	PEstado		novoEstado;

	assert( 0 < _numProposicoes );
	assert( NULL != _origem );

/*	cria o estado com todos os valores das proposicoes iguais a origem */

	novoEstado = ( PEstado )SM_MALLOC( "TEstado"
			, _numProposicoes * sizeof( TEstado ) );
	if( NULL == novoEstado )
		{
		fprintf( stderr, "Erro na alocacao de memoria\n" );
		return( NULL );
		}

	memcpy( novoEstado, _origem, _numProposicoes * sizeof( TEstado ) );

	return( novoEstado );
	}

/*
//	funcao para copiar um vetor de estado a partir de outro
*/

PEstado copiarEstado( long int _numProposicoes, PEstado _destino
		, PEstado _origem )
	{
	assert( 0 < _numProposicoes );
	assert( NULL != _destino );
	assert( NULL != _origem );

	memcpy( _destino, _origem, _numProposicoes * sizeof( TEstado ) );

	return( _destino );
	}

/*
//	funcao para destruir um vetor de estado
*/

void destruirEstado( PEstado _vetor )
	{
	assert( NULL != _vetor );

	SM_FREE( _vetor );
	}

/*
//	funcao para escrever um vetor de estado em um stream
*/

void escreverEstado( long int _numProposicoes, PProposicao _vetorProposicao
		, PEstado _vetorEstado, FILE * _stream )
	{
	long int	i = 0L;
	long int	j = 0L;

	assert( 0 < _numProposicoes );
	assert( NULL != _vetorProposicao );
	assert( NULL != _vetorEstado );
	assert( NULL != _stream );

	for( ; _numProposicoes > i; i++ )
		{
		if( INDEFINIDO == _vetorEstado[ i ].valorVerdade )
			continue;

		if( 0L < j )
			fprintf( _stream, " ^ " );

		if( VERDADEIRO == _vetorEstado[ i ].valorVerdade )
			fprintf( _stream, "%s", _vetorProposicao[ i ].nome );
		else
			fprintf( _stream, "~%s", _vetorProposicao[ i ].nome );

		j++;
		}

	fprintf( _stream, "\n" );
	}

/*
//	funcao para comparar dois estados - retorna 0 se forem iguais
*/

int compararEstados( long int _numProposicoes, PEstado _estado1
		, PEstado _estado2 )
	{
	long int	i = 0L;

	assert( 0 < _numProposicoes );
	assert( NULL != _estado1 );
	assert( NULL != _estado2 );

	for( ; _numProposicoes > i; i++ )
		{
		if( INDEFINIDO == _estado1[ i ].valorVerdade
				&& VERDADEIRO != _estado2[ i ].valorVerdade )
			continue;

		if( VERDADEIRO == _estado1[ i ].valorVerdade
				&& VERDADEIRO == _estado2[ i ].valorVerdade )
			continue;

		if( FALSO == _estado1[ i ].valorVerdade
				&& VERDADEIRO != _estado2[ i ].valorVerdade )
			continue;

		return( 1 );
		}

	return( 0 );
	}

/* /////////////////////////////////////////////////////////////////////////////
// operador STRIPS
///////////////////////////////////////////////////////////////////////////// */

/*
//	funcao para destruir um vetor de operadores
*/

void destruirOperador( long int _numOperadores, POperadorStrips _vetor )
	{
	long int	i = 0L;

	assert( 0 < _numOperadores );
	assert( NULL != _vetor );

	for( ; _numOperadores > i; i++ )
		{
		if( NULL != _vetor[ i ].precondicao )
			destruirEstado( _vetor[ i ].precondicao );

		if( NULL != _vetor[ i ].efeito )
			destruirEstado( _vetor[ i ].efeito );
		}

	SM_FREE( _vetor );
	}

/*
//	funcao para escrever a descricao de um operador em um stream
*/

void escreverOperador( long int _numProposicoes, PProposicao _vetorProposicao
		, POperadorStrips _operador, FILE * _stream )
	{
	assert( NULL != _operador );
	assert( NULL != _stream );

	fprintf( _stream, "operador: %s\n", _operador->nome );

	if( NULL != _operador->precondicao )
		{
		fprintf( _stream, "    pre-condicao: " );
		escreverEstado( _numProposicoes, _vetorProposicao, _operador->precondicao
				, _stream );
		}

	if( NULL != _operador->efeito )
		{
		fprintf( _stream, "    efeito: " );
		escreverEstado( _numProposicoes, _vetorProposicao, _operador->efeito
				, _stream );
		}
	}

/*
//	funcao para inserir um operador em um vetor de operadores
*/

POperadorStrips inserirOperador( long int _numOperadores, POperadorStrips _vetor
		, char * _nome )
	{
	POperadorStrips	novoOperador;

	assert( NULL != _nome );

	if( NULL == _vetor )
		{
		assert( 1 == _numOperadores );

		novoOperador = ( POperadorStrips )SM_MALLOC( "TOperadorStrips"
				, _numOperadores * sizeof( TOperadorStrips ) );
		if( NULL == novoOperador )
			{
			fprintf( stderr, "Erro na alocacao de memoria\n" );
			return( NULL );
			}
		}
	else
		{
		assert( 1 < _numOperadores );

		novoOperador = ( POperadorStrips )SM_REALLOC( "TOperadorStrips", _vetor
				, _numOperadores * sizeof( TOperadorStrips ) );
		if( NULL == novoOperador )
			{
			fprintf( stderr, "Erro na alocacao de memoria\n" );
			return( NULL );
			}
		}

	strcpy( novoOperador[ _numOperadores - 1 ].nome, _nome );
	novoOperador[ _numOperadores - 1 ].precondicao = NULL;
	novoOperador[ _numOperadores - 1 ].efeito = NULL;

	return( novoOperador );
	}

/*
//	funcao que retorna 1 se um operador eh aplicavel num estado
*/

int ehAplicavel( long int _numProposicoes, PEstado _vetorEstado
		, POperadorStrips _operador )
	{
	long int	i = 0L;

	assert( 0 < _numProposicoes );
	assert( NULL != _vetorEstado );
	assert( NULL != _operador );

	if( NULL == _operador->precondicao )
		return( 1 );

	for( ; _numProposicoes > i; i++ )
		{
		if( INDEFINIDO == _operador->precondicao[ i ].valorVerdade )
			continue;

		if( ( VERDADEIRO == _operador->precondicao[ i ].valorVerdade
				&& VERDADEIRO != _vetorEstado[ i ].valorVerdade )
				|| ( FALSO == _operador->precondicao[ i ].valorVerdade
				&& VERDADEIRO == _vetorEstado[ i ].valorVerdade ) )
			return( 0 );
		}

	return( 1 );
	}

/*
//	funcao que retorna o estado resultante da aplicacao de um operador num estado
*/

PEstado estadoResultante( long int _numProposicoes, PEstado _vetorEstado
		, POperadorStrips _operador )
	{
	PEstado		novoEstado;
	long int	i = 0L;

	assert( 0 < _numProposicoes );
	assert( NULL != _vetorEstado );
	assert( NULL != _operador );

/*	cria o estado com todos os valores das proposicoes iguais a origem */

	novoEstado = ( PEstado )SM_MALLOC( "TEstado"
			, _numProposicoes * sizeof( TEstado ) );
	if( NULL == novoEstado )
		{
		fprintf( stderr, "Erro na alocacao de memoria\n" );
		return( NULL );
		}

	memcpy( novoEstado, _vetorEstado, _numProposicoes * sizeof( TEstado ) );

/*	altera as proposicoes contidas no efeito do operador */

	for( ; _numProposicoes > i; i++ )
		{
		if( INDEFINIDO == _operador->efeito[ i ].valorVerdade )
			continue;

		novoEstado[ i ].valorVerdade = _operador->efeito[ i ].valorVerdade;
		}

	return( novoEstado );
	}

/* /////////////////////////////////////////////////////////////////////////////
// problema de planejamento STRIPS
///////////////////////////////////////////////////////////////////////////// */

/*
//	funcao para construir um vetor de estado
*/

void construirProblemaStrips( PProblemaStrips _problema )
	{
	assert( NULL != _problema );

	_problema->numProposicoes = 0L;
	_problema->proposicoes = NULL;
	_problema->numOperadores = 0L;
	_problema->dominio = NULL;
	_problema->estadoInicial = NULL;
	_problema->objetivo = NULL;
	}

/*
//	funcao para destruir um problema Strips
*/

void destruirProblemaStrips( PProblemaStrips _problema )
	{
	assert( NULL != _problema );

	destruirProposicoes( _problema->proposicoes );
	destruirOperador( _problema->numOperadores, _problema->dominio );
	destruirEstado( _problema->estadoInicial );
	destruirEstado( _problema->objetivo );
	}

/*
//	funcao para escrever um problema Strips em um stream
*/

void escreverProblemaStrips( PProblemaStrips _problema, FILE * _stream )
	{
	long int	i = 0L;

	assert( NULL != _problema );
	assert( NULL != _stream );

	fprintf( _stream, "proposicoes:\n" );
	escreverProposicoes( _problema->numProposicoes, _problema->proposicoes
		, _stream );

	fprintf( _stream, "descricao do dominio:\n" );
	for( ; _problema->numOperadores > i; i++ )
		{
		escreverOperador( _problema->numProposicoes, _problema->proposicoes
				, _problema->dominio + i, _stream );
		}

	fprintf( _stream, "estado inicial: " );
	escreverEstado( _problema->numProposicoes, _problema->proposicoes
		, _problema->estadoInicial, _stream );

	fprintf( _stream, "objetivo: " );
	escreverEstado( _problema->numProposicoes, _problema->proposicoes
		, _problema->objetivo, _stream );
	}

/*
//	funcao para adicionar em um problema Strips as proposicoes de uma linha
*/

PProposicao inserirProposicoesLinha( PProblemaStrips _problema, char * _linha )
	{
	PProposicao	novoVetor = NULL;
	char		proposicao[ 101 ];
	const char	delimitador = ';';
	char *		p = _linha;
	char *		q = _linha;

	assert( NULL != _linha );

	for( ;; )
		{
		if( '\0' == *q
				&& 0 == q - p )
			break;

		if( delimitador == *q
				|| '\0' == *q )
			{
			strncpy( proposicao, p, q - p );
			proposicao[ q - p ] = '\0';

			if( '~' != proposicao[ 0 ] )
				{
				if( -1 == procurarProposicao( _problema->numProposicoes
						, _problema->proposicoes, proposicao ) )
					{
					novoVetor = inserirProposicao( ++_problema->numProposicoes
							, _problema->proposicoes, proposicao );
					if( NULL == novoVetor )
						return( NULL );

					_problema->proposicoes = novoVetor;
					}
				}
			else
				{
				if( -1 == procurarProposicao( _problema->numProposicoes
						, _problema->proposicoes, proposicao + 1 ) )
					{
					novoVetor = inserirProposicao( ++_problema->numProposicoes
							, _problema->proposicoes, proposicao + 1 );
					if( NULL == novoVetor )
						return( NULL );

					_problema->proposicoes = novoVetor;
					}
				}

			if( '\0' == *q )
				break;

			p = q + 1;
			}

		q++;
		}

	return( _problema->proposicoes );
	}

/*
//	funcao para ler um problema de planejamento STRIPS
*/

int lerProblemaSTRIPS( FILE * _stream, PProblemaStrips _problema )
	{
	char		linha[ 16384 ];
	int			estado = 1;
	long int	i;

	assert( NULL != _stream );
	assert( NULL != _problema );

/*	le o stream para compor a lista de proposicoes */

	while( fgets( linha, sizeof( linha ), _stream ) != NULL )
		{
/*	elimina os caracteres de termino de linha */

		i = strlen( linha );
		while( 0 < i
				&& ( '\r' == linha[ i - 1 ]
				|| '\n' == linha[ i - 1 ]
				|| ' ' == linha[ i - 1 ] ) )
			{
			linha[ i - 1 ] = '\0';
			i--;
			}

/*	se eh uma linha valida, eh um nome de operador */

		switch( estado )
			{
		case 1:
			if( strcmp( linha, "" ) == 0 )
				{
				estado = 4;
				break;
				}

			estado = 2;
			break;

/*	a pre-condicao do ultimo de operador */

		case 2:
			if( '\0' != linha[ 0 ]
					&& NULL == inserirProposicoesLinha( _problema, linha ) )
				return( -1 );

			estado = 3;
			break;

/*	o efeito do ultimo de operador */

		case 3:
			if( '\0' != linha[ 0 ]
					&& NULL == inserirProposicoesLinha( _problema, linha ) )
				return( -1 );

			estado = 1;
			break;

/*	o estado inicial */

		case 4:
			if( '\0' != linha[ 0 ]
					&& NULL == inserirProposicoesLinha( _problema, linha ) )
				return( -1 );

			estado = 5;
			break;

/*	o objetivo */

		case 5:
			if( '\0' != linha[ 0 ]
					&& NULL == inserirProposicoesLinha( _problema, linha ) )
				return( -1 );

			estado = 6;
			break;

		default:
			break;
			}
		}

/*	agora sim, faz a leitura do dominio e dos estados inicial e objetivo */

	estado = 1;

	fseek( _stream, 0L, SEEK_SET );

	while( fgets( linha, sizeof( linha ), _stream ) != NULL )
		{
/*	elimina os caracteres de termino de linha */

		i = strlen( linha );
		while( 0 < i
				&& ( '\r' == linha[ i - 1 ]
				|| '\n' == linha[ i - 1 ]
				|| ' ' == linha[ i - 1 ] ) )
			{
			linha[ i - 1 ] = '\0';
			i--;
			}

/*	se eh uma linha valida, eh um nome de operador */

		switch( estado )
			{
		case 1:
			if( strcmp( linha, "" ) == 0 )
				{
				estado = 4;
				break;
				}

			_problema->dominio = inserirOperador( ++_problema->numOperadores
					, _problema->dominio, linha );

			estado = 2;
			break;

/*	a pre-condicao do ultimo de operador */

		case 2:
			_problema->dominio[ _problema->numOperadores - 1 ].precondicao = construirEstado( _problema->numProposicoes
					, _problema->proposicoes, linha );

			estado = 3;
			break;

/*	o efeito do ultimo de operador */

		case 3:
			_problema->dominio[ _problema->numOperadores - 1 ].efeito = construirEstado( _problema->numProposicoes
					, _problema->proposicoes, linha );

			estado = 1;
			break;

/*	o estado inicial */

		case 4:
			_problema->estadoInicial = construirEstado( _problema->numProposicoes
					, _problema->proposicoes, linha );

			estado = 5;
			break;

/*	o objetivo */

		case 5:
			_problema->objetivo = construirEstado( _problema->numProposicoes
					, _problema->proposicoes, linha );

			estado = 6;
			break;

		default:
			break;
			}
		}

	return( 0 );
	}

