/* /////////////////////////////////////////////////////////////////////////////
// strips.h  -  19/05/2002 by aldeba
//
// Estruturas de dados e prototipos das funcoes para a leitura de um problema
//    de planejamento STRIPS
///////////////////////////////////////////////////////////////////////////// */

/*	protecao contra referencias circulares */

#if			!defined( __STRIPS_H )

#define		__STRIPS_H

/*	includes do fonte */

#include	<stdio.h>

/* /////////////////////////////////////////////////////////////////////////////
// declaracoes de estruturas de dados
///////////////////////////////////////////////////////////////////////////// */

// proposicoes

struct proposicao {
	char						nome[ 101 ];
	long int					numProposicao;
};

typedef struct proposicao		TProposicao;
typedef TProposicao *			PProposicao;

// estado

#define	VERDADEIRO	 1
#define	INDEFINIDO	 0
#define	FALSO		-1

struct estado {
	short int					valorVerdade;
};

typedef struct estado			TEstado;
typedef TEstado *				PEstado;

// operador STRIPS

struct operadorStrips {
	char						nome[ 101 ];
	PEstado						precondicao;
	PEstado						efeito;
};

typedef struct operadorStrips	TOperadorStrips;
typedef TOperadorStrips *		POperadorStrips;

// problema de planejamento STRIPS

struct problemaStrips {
	long int					numProposicoes;
	PProposicao					proposicoes;
	long int					numOperadores;
	POperadorStrips				dominio;
	PEstado						estadoInicial;
	PEstado						objetivo;
};

typedef struct problemaStrips	TProblemaStrips;
typedef TProblemaStrips *		PProblemaStrips;

/* /////////////////////////////////////////////////////////////////////////////
// declaracoes de prototipos
///////////////////////////////////////////////////////////////////////////// */

void			destruirProposicoes		( PProposicao _vetor );
int				compararProposicoes		( const void * _p1, const void * _p2 );
void			escreverProposicoes		( long int _numProposicoes, PProposicao _vetor, FILE * _stream );
PProposicao		inserirProposicao		( long int _numProposicoes, PProposicao _vetor, char * _proposicao );
long int		procurarProposicao		( long int _numProposicoes, PProposicao _vetor, char * _proposicao );

PEstado			construirEstado			( long int _numProposicoes, PProposicao _vetor, char * _linha );
PEstado			construirCopiaEstado	( long int _numProposicoes, PEstado _origem );
PEstado			copiarEstado			( long int _numProposicoes, PEstado _destino, PEstado _origem );
void			destruirEstado			( PEstado _vetor );
void			escreverEstado			( long int _numProposicoes, PProposicao _vetorProposicao, PEstado _vetorEstado, FILE * _stream );
int				compararEstados			( long int _numProposicoes, PEstado _estado1, PEstado _estado2 );

void			destruirOperador		( long int _numOperadores, POperadorStrips _vetor );
void			escreverOperador		( long int _numProposicoes, PProposicao _vetorProposicao, POperadorStrips _operador, FILE * _stream );
POperadorStrips	inserirOperador			( long int _numOperadores, POperadorStrips _vetor, char * _nome );
int				ehAplicavel				( long int _numProposicoes, PEstado _vetorEstado, POperadorStrips _operador );
PEstado			estadoResultante		( long int _numProposicoes, PEstado _vetorEstado, POperadorStrips _operador );

void			construirProblemaStrips	( PProblemaStrips _problema );
void			destruirProblemaStrips	( PProblemaStrips _problema );
void			escreverProblemaStrips	( PProblemaStrips _problema, FILE * _stream );
PProposicao		inserirProposicoesLinha	( PProblemaStrips _problema, char * _linha );
int				lerProblemaSTRIPS		( FILE * _stream, PProblemaStrips _problema );
#endif		/* !defined( __STRIPS_H ) */

